@@ -26,10 +26,7 @@
 
 #include <linux/bitops.h>
 #include <linux/vmalloc.h>
-#include <linux/string.h> // for memset
-#include <linux/hardirq.h> /* for D_ASSERT(in_interrupt()) */
-
-
+#include <linux/string.h>
 #include <linux/drbd.h>
 #include "drbd_int.h"
 
@@ -88,120 +85,97 @@ struct drbd_bitmap {
 	 * rather change the granularity to 16k or 64k or something.
 	 * (that implies other problems, however...)
 	 */
-	unsigned long bm_fo;        // next offset for drbd_bm_find_next
-	unsigned long bm_set;       // nr of set bits; THINK maybe atomic_t ?
+	unsigned long bm_fo;        /* next offset for drbd_bm_find_next */
+	unsigned long bm_set;       /* nr of set bits; THINK maybe atomic_t? */
 	unsigned long bm_bits;
 	size_t   bm_words;
 	sector_t bm_dev_capacity;
-	struct semaphore bm_change; // serializes resize operations
+	struct semaphore bm_change; /* serializes resize operations */
 
 	atomic_t bm_async_io;
 	wait_queue_head_t bm_io_wait;
 
 	unsigned long  bm_flags;
 
-	// { REMOVE
-	unsigned long  bm_line;
-	char          *bm_file;
-	// }
+	/* debugging aid, in case we are still racy somewhere */
+	char          *bm_why;
+	struct task_struct *bm_task;
 };
 
-// { REMOVE once we serialize all state changes properly
-#define D_BUG_ON(x)	ERR_IF(x) { dump_stack(); }
+/* definition of bits in bm_flags */
 #define BM_LOCKED 0
-#define BM_MD_IO_ERROR (BITS_PER_LONG-1) // 31? 63?
+#define BM_MD_IO_ERROR (BITS_PER_LONG-1) /* 31? 63? */
 
-#if 0 // simply disabled for now...
-#define MUST_NOT_BE_LOCKED() do {					\
-	if (test_bit(BM_LOCKED,&b->bm_flags)) {				\
-		if (DRBD_ratelimit(5*HZ,5)) {				\
-			ERR("%s:%d: bitmap is locked by %s:%lu\n",	\
-			    __FILE__, __LINE__, b->bm_file,b->bm_line);	\
-			dump_stack();					\
-		}							\
-	}								\
-} while (0)
-#define MUST_BE_LOCKED() do {						\
-	if (!test_bit(BM_LOCKED,&b->bm_flags)) {			\
-		if (DRBD_ratelimit(5*HZ,5)) {				\
-			ERR("%s:%d: bitmap not locked!\n",		\
-					__FILE__, __LINE__);		\
-			dump_stack();					\
-		}							\
-	}								\
-} while (0)
-#else
-#define MUST_NOT_BE_LOCKED() do {(void)b;} while (0)
-#define MUST_BE_LOCKED() do {(void)b;} while (0)
-#endif
-void __drbd_bm_lock(drbd_dev *mdev, char* file, int line)
+static inline int bm_is_locked(struct drbd_bitmap *b)
 {
-	struct drbd_bitmap *b = mdev->bitmap;
-
-	D_BUG_ON(irqs_disabled());
-	spin_lock_irq(&b->bm_lock);
-	if (!__test_and_set_bit(BM_LOCKED,&b->bm_flags)) {
-		b->bm_file = file;
-		b->bm_line = line;
-	} else if (DRBD_ratelimit(5*HZ,5)) {
-		ERR("%s:%d: bitmap already locked by %s:%lu\n",
-		    file, line, b->bm_file,b->bm_line);
-		/*
-		dump_stack();
-		ERR("This is no oops, but debug stack trace only.\n");
-		ERR("If you get this often, or in reproducable situations, "
-		    "notify <drbd-devel@linbit.com>\n");
-		*/
-	}
-	spin_unlock_irq(&b->bm_lock);
-}
-void drbd_bm_unlock(drbd_dev *mdev)
-{
-	struct drbd_bitmap *b = mdev->bitmap;
-	D_BUG_ON(irqs_disabled());
-	spin_lock_irq(&b->bm_lock);
-	if (!__test_and_clear_bit(BM_LOCKED,&mdev->bitmap->bm_flags)) {
-		ERR("bitmap not locked in bm_unlock\n");
-	} else {
-		/* FIXME if we got a "is already locked" previously,
-		 * we unlock here even though we actually MUST NOT do so... */
-		b->bm_file = NULL;
-		b->bm_line = -1;
-	}
-	spin_unlock_irq(&b->bm_lock);
+	return test_bit(BM_LOCKED, &b->bm_flags);
 }
 
-#if 0
-// has been very helpful to indicate that rs_total and rs_left have been
-// used in a non-smp safe way...
-#define BM_PARANOIA_CHECK() do {						\
-	D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);				\
-	D_ASSERT(b->bm_dev_capacity == drbd_get_capacity(mdev->this_bdev));	\
-	if ( (b->bm_set != mdev->rs_total) &&					\
-	     (b->bm_set != mdev->rs_left) ) {					\
-		if ( DRBD_ratelimit(5*HZ,5) ) {					\
-			ERR("%s:%d: ?? bm_set=%lu; rs_total=%lu, rs_left=%lu\n",\
-				__FILE__ , __LINE__ ,				\
-				b->bm_set, mdev->rs_total, mdev->rs_left );	\
-		}								\
-	}									\
-} while (0)
-#else
-#define BM_PARANOIA_CHECK() do {					\
-	D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);			\
-	if (b->bm_dev_capacity != drbd_get_capacity(mdev->this_bdev)) {	\
-		ERR("%s:%d: bm_dev_capacity:%llu drbd_get_capacity:%llu\n", \
-		__FILE__, __LINE__,					\
-		(unsigned long long) b->bm_dev_capacity,		\
-		(unsigned long long) drbd_get_capacity(mdev->this_bdev));\
-	}								\
-} while (0)
-#endif
-// }
+#define bm_print_lock_info(m) __bm_print_lock_info(m, __func__)
+static void __bm_print_lock_info(struct drbd_conf *mdev, const char *func)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	if (!DRBD_ratelimit(5*HZ, 5))
+		return;
+	ERR("FIXME %s in %s, bitmap locked for '%s' by %s\n",
+	    current == mdev->receiver.task ? "receiver" :
+	    current == mdev->asender.task  ? "asender"  :
+	    current == mdev->worker.task   ? "worker"   : current->comm,
+	    func, b->bm_why ?: "?",
+	    b->bm_task == mdev->receiver.task ? "receiver" :
+	    b->bm_task == mdev->asender.task  ? "asender"  :
+	    b->bm_task == mdev->worker.task   ? "worker"   : "?");
+}
+
+void drbd_bm_lock(struct drbd_conf *mdev, char *why)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	int trylock_failed;
+
+	if (!b) {
+		ERR("FIXME no bitmap in drbd_bm_lock!?\n");
+		return;
+	}
+
+	trylock_failed = down_trylock(&b->bm_change);
+
+	if (trylock_failed) {
+		DBG("%s going to '%s' but bitmap already locked for '%s' by %s\n",
+		    current == mdev->receiver.task ? "receiver" :
+		    current == mdev->asender.task  ? "asender"  :
+		    current == mdev->worker.task   ? "worker"   : "?",
+		    why, b->bm_why ?: "?",
+		    b->bm_task == mdev->receiver.task ? "receiver" :
+		    b->bm_task == mdev->asender.task  ? "asender"  :
+		    b->bm_task == mdev->worker.task   ? "worker"   : "?");
+		down(&b->bm_change);
+	}
+	if (__test_and_set_bit(BM_LOCKED,&b->bm_flags))
+		ERR("FIXME bitmap already locked in bm_lock\n");
+
+	b->bm_why  = why;
+	b->bm_task = current;
+}
+
+void drbd_bm_unlock(struct drbd_conf *mdev)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	if (!b) {
+		ERR("FIXME no bitmap in drbd_bm_unlock!?\n");
+		return;
+	}
+
+	if (!__test_and_clear_bit(BM_LOCKED, &mdev->bitmap->bm_flags))
+		ERR("FIXME bitmap not locked in bm_unlock\n");
+
+	b->bm_why  = NULL;
+	b->bm_task = NULL;
+	up(&b->bm_change);
+}
 
 #if DUMP_MD >= 3
 /* debugging aid */
-STATIC void bm_end_info(drbd_dev *mdev, const char* where)
+STATIC void bm_end_info(struct drbd_conf *mdev, const char *where)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	size_t w = (b->bm_bits-1) >> LN2_BPL;
@@ -210,9 +184,9 @@ STATIC void bm_end_info(drbd_dev *mdev, const char* where)
 	INFO("bm[%d]=0x%lX\n", w, b->bm[w]);
 	w++;
 
-	if ( w < b->bm_words ) {
-		D_ASSERT(w == b->bm_words -1);
-		INFO("bm[%d]=0x%lX\n",w,b->bm[w]);
+	if (w < b->bm_words) {
+		D_ASSERT(w == b->bm_words - 1);
+		INFO("bm[%d]=0x%lX\n", w, b->bm[w]);
 	}
 }
 #else
@@ -224,7 +198,7 @@ STATIC void bm_end_info(drbd_dev *mdev, const char* where)
 
 /*
  * actually most functions herein should take a struct drbd_bitmap*, not a
- * drbd_dev*, but for the debug macros I like to have the mdev around
+ * struct drbd_conf*, but for the debug macros I like to have the mdev around
  * to be able to report device specific.
  */
 
@@ -238,11 +212,11 @@ STATIC void bm_end_info(drbd_dev *mdev, const char* where)
  * called on driver init only. TODO call when a device is created.
  * allocates the drbd_bitmap, and stores it in mdev->bitmap.
  */
-int drbd_bm_init(drbd_dev *mdev)
+int drbd_bm_init(struct drbd_conf *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	D_BUG_ON(b);
-	b = kzalloc(sizeof(struct drbd_bitmap),GFP_KERNEL);
+	WARN_ON(b != NULL);
+	b = kzalloc(sizeof(struct drbd_bitmap), GFP_KERNEL);
 	if (!b)
 		return -ENOMEM;
 	spin_lock_init(&b->bm_lock);
@@ -254,7 +228,7 @@ int drbd_bm_init(drbd_dev *mdev)
 	return 0;
 }
 
-sector_t drbd_bm_capacity(drbd_dev *mdev)
+sector_t drbd_bm_capacity(struct drbd_conf *mdev)
 {
 	ERR_IF(!mdev->bitmap) return 0;
 	return mdev->bitmap->bm_dev_capacity;
@@ -262,13 +236,13 @@ sector_t drbd_bm_capacity(drbd_dev *mdev)
 
 /* called on driver unload. TODO: call when a device is destroyed.
  */
-void drbd_bm_cleanup(drbd_dev *mdev)
+void drbd_bm_cleanup(struct drbd_conf *mdev)
 {
 	ERR_IF (!mdev->bitmap) return;
 	/* FIXME I think we should explicitly change the device size to zero
 	 * before this...
 	 *
-	D_BUG_ON(mdev->bitmap->bm);
+	WARN_ON(mdev->bitmap->bm);
 	 */
 	vfree(mdev->bitmap->bm);
 	kfree(mdev->bitmap);
@@ -280,37 +254,35 @@ void drbd_bm_cleanup(drbd_dev *mdev)
  * this masks out the remaining bits.
  * Rerturns the number of bits cleared.
  */
-STATIC int bm_clear_surplus(struct drbd_bitmap * b)
+STATIC int bm_clear_surplus(struct drbd_bitmap *b)
 {
-	const unsigned long mask = (1UL << (b->bm_bits & (BITS_PER_LONG-1))) -1;
+	const unsigned long mask = (1UL << (b->bm_bits & (BITS_PER_LONG-1))) - 1;
 	size_t w = b->bm_bits >> LN2_BPL;
-	int cleared=0;
+	int cleared = 0;
 
-	if ( w < b->bm_words ) {
+	if (w < b->bm_words) {
 		cleared = hweight_long(b->bm[w] & ~mask);
 		b->bm[w++] &= mask;
 	}
 
-	if ( w < b->bm_words ) {
+	if (w < b->bm_words) {
 		cleared += hweight_long(b->bm[w]);
-		b->bm[w++]=0;
+		b->bm[w++] = 0;
 	}
 
 	return cleared;
 }
 
-STATIC void bm_set_surplus(struct drbd_bitmap * b)
+STATIC void bm_set_surplus(struct drbd_bitmap *b)
 {
-	const unsigned long mask = (1UL << (b->bm_bits & (BITS_PER_LONG-1))) -1;
+	const unsigned long mask = (1UL << (b->bm_bits & (BITS_PER_LONG-1))) - 1;
 	size_t w = b->bm_bits >> LN2_BPL;
 
-	if ( w < b->bm_words ) {
+	if (w < b->bm_words)
 		b->bm[w++] |= ~mask;
-	}
 
-	if ( w < b->bm_words ) {
+	if (w < b->bm_words)
 		b->bm[w++] = ~(0UL);
-	}
 }
 
 STATIC unsigned long __bm_count_bits(struct drbd_bitmap *b, const int swap_endian)
@@ -319,9 +291,10 @@ STATIC unsigned long __bm_count_bits(struct drbd_bitmap *b, const int swap_endia
 	unsigned long *ep = b->bm + b->bm_words;
 	unsigned long bits = 0;
 
-	while ( bm < ep ) {
+	while (bm < ep) {
 #ifndef __LITTLE_ENDIAN
-		if (swap_endian) *bm = lel_to_cpu(*bm);
+		if (swap_endian)
+			*bm = lel_to_cpu(*bm);
 #endif
 		bits += hweight_long(*bm++);
 	}
@@ -331,30 +304,36 @@ STATIC unsigned long __bm_count_bits(struct drbd_bitmap *b, const int swap_endia
 
 static inline unsigned long bm_count_bits(struct drbd_bitmap *b)
 {
-	return __bm_count_bits(b,0);
+	return __bm_count_bits(b, 0);
 }
 
 static inline unsigned long bm_count_bits_swap_endian(struct drbd_bitmap *b)
 {
-	return __bm_count_bits(b,1);
+	return __bm_count_bits(b, 1);
 }
 
 
-void _drbd_bm_recount_bits(drbd_dev *mdev, char* file, int line)
+void _drbd_bm_recount_bits(struct drbd_conf *mdev, char *file, int line)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long flags, bits;
 
 	ERR_IF(!b) return;
 
-	spin_lock_irqsave(&b->bm_lock,flags);
+	/* IMO this should be inside drbd_bm_lock/unlock.
+	 * Unfortunately it is used outside of the locks.
+	 * And I'm not yet sure where we need to place the
+	 * lock/unlock correctly.
+	 */
+
+	spin_lock_irqsave(&b->bm_lock, flags);
 	bits = bm_count_bits(b);
-	if(bits != b->bm_set) {
+	if (bits != b->bm_set) {
 		ERR("bm_set was %lu, corrected to %lu. %s:%d\n",
-		    b->bm_set,bits,file,line);
+		    b->bm_set, bits, file, line);
 		b->bm_set = bits;
 	}
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irqrestore(&b->bm_lock, flags);
 }
 
 #define BM_SECTORS_PER_BIT (BM_BLOCK_SIZE/512)
@@ -367,18 +346,15 @@ void _drbd_bm_recount_bits(drbd_dev *mdev, char* file, int line)
  * In case this is actually a resize, we copy the old bitmap into the new one.
  * Otherwise, the bitmap is initiallized to all bits set.
  */
-int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
+int drbd_bm_resize(struct drbd_conf *mdev, sector_t capacity)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	unsigned long bits, bytes, words, *nbm, *obm = 0;
+	unsigned long bits, bytes, words, *nbm, *obm = NULL;
 	int err = 0, growing;
 
 	ERR_IF(!b) return -ENOMEM;
-	MUST_BE_LOCKED();
 
-	ERR_IF (down_trylock(&b->bm_change)) {
-		down(&b->bm_change);
-	}
+	drbd_bm_lock(mdev, "resize");
 
 	INFO("drbd_bm_resize called with capacity == %llu\n",
 			(unsigned long long)capacity);
@@ -398,71 +374,83 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		spin_unlock_irq(&b->bm_lock);
 		goto free_obm;
 	} else {
-		bits = BM_SECT_TO_BIT(ALIGN(capacity,BM_SECTORS_PER_BIT));
+		bits = BM_SECT_TO_BIT(ALIGN(capacity, BM_SECTORS_PER_BIT));
 
 		/* if we would use
 		   words = ALIGN(bits,BITS_PER_LONG) >> LN2_BPL;
 		   a 32bit host could present the wrong number of words
 		   to a 64bit host.
 		*/
-		words = ALIGN(bits,64) >> LN2_BPL;
+		words = ALIGN(bits, 64) >> LN2_BPL;
 
-		D_ASSERT((u64)bits <= (((u64)mdev->bc->md.md_size_sect-MD_BM_OFFSET) << 12));
+		if (inc_local(mdev)) {
+			D_ASSERT((u64)bits <= (((u64)mdev->bc->md.md_size_sect-MD_BM_OFFSET) << 12));
+			dec_local(mdev);
+		}
 
-		if ( words == b->bm_words ) {
+		growing = bits > b->bm_bits;
+		if (words == b->bm_words) {
 			/* optimize: capacity has changed,
-			 * but only within one long word worth of bits.
-			 * just update the bm_dev_capacity and bm_bits members.
-			 */
+			 * but only within one ulong64 word worth of bits.
+			 * no allocation needed, just update the
+			 * bm_dev_capacity and bm_bits members,
+			 * and set the new bits, if any */
 			spin_lock_irq(&b->bm_lock);
-			b->bm_bits    = bits;
-			b->bm_dev_capacity = capacity;
-			b->bm_set -= bm_clear_surplus(b);
-			bm_end_info(mdev, __FUNCTION__ );
-			spin_unlock_irq(&b->bm_lock);
-			goto out;
+			bm_set_surplus(b);
+			if (growing)
+				b->bm_set += bits - b->bm_bits;
+			goto done;
 		} else {
-		        /* one extra long to catch off by one errors */
+			/* one extra long to catch off by one errors */
 			bytes = (words+1)*sizeof(long);
 			nbm = vmalloc(bytes);
 			if (!nbm) {
-				ERR("bitmap: failed to vmalloc %lu bytes\n",bytes);
+				ERR("bitmap: failed to vmalloc %lu bytes\n",
+					bytes);
 				err = -ENOMEM;
 				goto out;
 			}
 		}
 		spin_lock_irq(&b->bm_lock);
 		obm = b->bm;
-		// brgs. move several MB within spinlock...
-		// FIXME this should go into userspace!
+		/* brgs. move several MB within spinlock...
+		 * FIXME this should go into userspace! */
 		if (obm) {
+			/* adjust for possibly partially used
+			 * last word of old bitmap. */
 			bm_set_surplus(b);
 			D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);
-			memcpy(nbm,obm,min_t(size_t,b->bm_words,words)*sizeof(long));
+			memcpy(nbm, obm, min_t(size_t, b->bm_words, words)
+								*sizeof(long));
 		}
-		growing = words > b->bm_words;
-		if (growing) { // set all newly allocated bits
-			// start at -1, just to be sure.
-			memset( nbm + (b->bm_words?:1)-1 , 0xff,
-				(words - ((b->bm_words?:1)-1)) * sizeof(long) );
+		if (growing) {
+			/* set all newly allocated bits */
+			memset(nbm + b->bm_words, 0xff,
+				(words - b->bm_words) * sizeof(long));
+			/* yes, I know, this is not the same number as was set by this memset.
+			 * bm_set_surplus above before the memcpy,
+			 * and bm_clear_surplus below after the new assignments
+			 * make sure that this is indeed the amount of newly set bits */
 			b->bm_set  += bits - b->bm_bits;
 		}
 		nbm[words] = DRBD_MAGIC;
 		b->bm = nbm;
-		b->bm_bits  = bits;
 		b->bm_words = words;
+ done:
+		b->bm_bits = bits;
 		b->bm_dev_capacity = capacity;
+		/* finally clear possibly only partially used last words */
 		bm_clear_surplus(b);
 		if (!growing)
 			b->bm_set = bm_count_bits(b);
-		bm_end_info(mdev, __FUNCTION__ );
+		bm_end_info(mdev, __func__);
 		spin_unlock_irq(&b->bm_lock);
-		INFO("resync bitmap: bits=%lu words=%lu\n",bits,words);
+		INFO("resync bitmap: bits=%lu words=%lu\n", bits, words);
 	}
  free_obm:
-	vfree(obm); // vfree(NULL) is noop
+	vfree(obm); /* vfree(NULL) is noop */
  out:
-	up(&b->bm_change);
+	drbd_bm_unlock(mdev);
 	return err;
 }
 
@@ -474,61 +462,51 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
  *
  * maybe bm_set should be atomic_t ?
  */
-unsigned long drbd_bm_total_weight(drbd_dev *mdev)
+unsigned long drbd_bm_total_weight(struct drbd_conf *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long s;
 	unsigned long flags;
 
 	ERR_IF(!b) return 0;
-	// MUST_BE_LOCKED(); well. yes. but ...
 
-	spin_lock_irqsave(&b->bm_lock,flags);
+	spin_lock_irqsave(&b->bm_lock, flags);
 	s = b->bm_set;
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irqrestore(&b->bm_lock, flags);
 
 	return s;
 }
 
-size_t drbd_bm_words(drbd_dev *mdev)
+size_t drbd_bm_words(struct drbd_conf *mdev)
 {
-	struct drbd_bitmap *b = mdev->bitmap;
-	ERR_IF(!b) return 0;
-
-	/* FIXME
-	 * actually yes. really. otherwise it could just change its size ...
-	 * but it triggers all the time...
-	 * MUST_BE_LOCKED();
-	 */
-
-	return b->bm_words;
+	ERR_IF(!mdev->bitmap) return 0;
+	return mdev->bitmap->bm_words;
 }
 
 /* merge number words from buffer into the bitmap starting at offset.
  * buffer[i] is expected to be little endian unsigned long.
+ * bitmap must be locked by drbd_bm_lock.
+ * currently only used from receive_bitmap.
  */
-void drbd_bm_merge_lel( drbd_dev *mdev, size_t offset, size_t number,
-			unsigned long* buffer )
+void drbd_bm_merge_lel(struct drbd_conf *mdev, size_t offset, size_t number,
+			unsigned long *buffer)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long *bm;
 	unsigned long word, bits;
 	size_t n = number;
 
-	if (number == 0) return;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
-	D_BUG_ON(offset        >= b->bm_words);
-	D_BUG_ON(offset+number >  b->bm_words);
-	D_BUG_ON(number > PAGE_SIZE/sizeof(long));
+	if (number == 0)
+		return;
+	WARN_ON(offset        >= b->bm_words);
+	WARN_ON(offset+number >  b->bm_words);
+	WARN_ON(number > PAGE_SIZE/sizeof(long));
 
-	MUST_BE_LOCKED();
-
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	// BM_PARANOIA_CHECK(); no.
 	bm = b->bm + offset;
-	while(n--) {
+	while (n--) {
 		bits = hweight_long(*bm);
 		word = *bm | lel_to_cpu(*buffer++);
 		*bm++ = word;
@@ -541,7 +519,7 @@ void drbd_bm_merge_lel( drbd_dev *mdev, size_t offset, size_t number,
 	 */
 	if (offset+number == b->bm_words) {
 		b->bm_set -= bm_clear_surplus(b);
-		bm_end_info(mdev, __FUNCTION__ );
+		bm_end_info(mdev, __func__);
 	}
 	spin_unlock_irq(&b->bm_lock);
 }
@@ -549,76 +527,65 @@ void drbd_bm_merge_lel( drbd_dev *mdev, size_t offset, size_t number,
 /* copy number words from the bitmap starting at offset into the buffer.
  * buffer[i] will be little endian unsigned long.
  */
-void drbd_bm_get_lel( drbd_dev *mdev, size_t offset, size_t number,
-		      unsigned long* buffer )
+void drbd_bm_get_lel(struct drbd_conf *mdev, size_t offset, size_t number,
+		     unsigned long *buffer)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long *bm;
 
-	if (number == 0) return;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
-	if ( (offset        >= b->bm_words) ||
-	     (offset+number >  b->bm_words) ||
-	     (number > PAGE_SIZE/sizeof(long)) ||
-	     (number <= 0) ) {
-		// yes, there is "%z", but that gives compiler warnings...
+	spin_lock_irq(&b->bm_lock);
+	if ((offset        >= b->bm_words) ||
+	    (offset+number >  b->bm_words) ||
+	    (number > PAGE_SIZE/sizeof(long)) ||
+	    (number <= 0))
 		ERR("offset=%lu number=%lu bm_words=%lu\n",
 			(unsigned long)	offset,
 			(unsigned long)	number,
 			(unsigned long) b->bm_words);
-		return;
+	else {
+		bm = b->bm + offset;
+		while (number--)
+			*buffer++ = cpu_to_lel(*bm++);
 	}
-
-	// MUST_BE_LOCKED(); yes. but not neccessarily globally...
-
-	D_BUG_ON(irqs_disabled());
-	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
-	bm = b->bm + offset;
-	while(number--) *buffer++ = cpu_to_lel(*bm++);
 	spin_unlock_irq(&b->bm_lock);
 }
 
 /* set all bits in the bitmap */
-void drbd_bm_set_all(drbd_dev *mdev)
+void drbd_bm_set_all(struct drbd_conf *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
 
-	MUST_BE_LOCKED();
-
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
-	memset(b->bm,0xff,b->bm_words*sizeof(long));
+	memset(b->bm, 0xff, b->bm_words*sizeof(long));
 	bm_clear_surplus(b);
 	b->bm_set = b->bm_bits;
 	spin_unlock_irq(&b->bm_lock);
 }
 
-BIO_ENDIO_FN(bm_async_io_complete)
+static BIO_ENDIO_TYPE bm_async_io_complete BIO_ENDIO_ARGS(struct bio *bio, int error)
 {
 	struct drbd_bitmap *b = bio->bi_private;
-	int uptodate = bio_flagged(bio,BIO_UPTODATE);
+	int uptodate = bio_flagged(bio, BIO_UPTODATE);
 
 	BIO_ENDIO_FN_START;
 
-	if (!error && !uptodate) {
-		/* strange behaviour of some lower level drivers...
-		 * fail the request by clearing the uptodate flag,
-		 * but do not return any error?!
-		 * do we want to WARN() on this? */
+	/* strange behaviour of some lower level drivers...
+	 * fail the request by clearing the uptodate flag,
+	 * but do not return any error?!
+	 * do we want to WARN() on this? */
+	if (!error && !uptodate)
 		error = -EIO;
-	}
 
 	if (error) {
 		/* doh. what now?
 		 * for now, set all bits, and flag MD_IO_ERROR
 		 */
 		/* FIXME kmap_atomic memset etc. pp. */
-		__set_bit(BM_MD_IO_ERROR,&b->bm_flags);
+		__set_bit(BM_MD_IO_ERROR, &b->bm_flags);
 	}
 	if (atomic_dec_and_test(&b->bm_async_io))
 		wake_up(&b->bm_io_wait);
@@ -628,17 +595,20 @@ BIO_ENDIO_FN(bm_async_io_complete)
 	BIO_ENDIO_FN_RETURN;
 }
 
-STATIC void bm_page_io_async(drbd_dev *mdev, struct drbd_bitmap *b, int page_nr, int rw)
+STATIC void bm_page_io_async(struct drbd_conf *mdev, struct drbd_bitmap *b, int page_nr, int rw) __must_hold(local)
 {
 	/* we are process context. we always get a bio */
 	/* THINK: do we need GFP_NOIO here? */
 	struct bio *bio = bio_alloc(GFP_KERNEL, 1);
-	struct page *page = vmalloc_to_page((char*)(b->bm) + (PAGE_SIZE*page_nr));
+	struct page *page = vmalloc_to_page((char *)(b->bm)
+						+ (PAGE_SIZE*page_nr));
 	unsigned int len;
-	sector_t on_disk_sector = mdev->bc->md.md_offset + mdev->bc->md.bm_offset;
+	sector_t on_disk_sector =
+		mdev->bc->md.md_offset + mdev->bc->md.bm_offset;
 	on_disk_sector += ((sector_t)page_nr) << (PAGE_SHIFT-9);
 
-	/* this might happen with very small flexible external meta data device */
+	/* this might happen with very small
+	 * flexible external meta data device */
 	len = min_t(unsigned int, PAGE_SIZE,
 		(drbd_md_last_sector(mdev->bc) - on_disk_sector + 1)<<9);
 
@@ -651,12 +621,12 @@ STATIC void bm_page_io_async(drbd_dev *mdev, struct drbd_bitmap *b, int page_nr,
 	bio->bi_private = b;
 	bio->bi_end_io = bm_async_io_complete;
 
-	if (FAULT_ACTIVE(mdev, (rw&WRITE)?DRBD_FAULT_MD_WR:DRBD_FAULT_MD_RD)) {
+	if (FAULT_ACTIVE(mdev, (rw & WRITE) ? DRBD_FAULT_MD_WR : DRBD_FAULT_MD_RD)) {
 		bio->bi_rw |= rw;
 		bio_endio(bio, -EIO);
-	}
-	else
+	} else {
 		submit_bio(rw, bio);
+	}
 }
 
 # if defined(__LITTLE_ENDIAN)
@@ -669,7 +639,7 @@ void bm_cpu_to_lel(struct drbd_bitmap *b)
 	 * this may be optimized by using
 	 * cpu_to_lel(-1) == -1 and cpu_to_lel(0) == 0;
 	 * the following is still not optimal, but better than nothing */
-	const unsigned long *end = b->bm+b->bm_words;
+	unsigned long *end = b->bm+b->bm_words;
 	unsigned long *bm;
 	if (b->bm_set == 0) {
 		/* no page at all; avoid swap if all is 0 */
@@ -681,9 +651,8 @@ void bm_cpu_to_lel(struct drbd_bitmap *b)
 		/* all pages */
 		bm = b->bm;
 	}
-	for (; bm < end; bm++) {
+	for (; bm < end; bm++)
 		*bm = cpu_to_lel(*bm);
-	}
 }
 # endif
 /* lel_to_cpu == cpu_to_lel */
@@ -692,7 +661,7 @@ void bm_cpu_to_lel(struct drbd_bitmap *b)
 /*
  * bm_rw: read/write the whole bitmap from/to its on disk location.
  */
-STATIC int bm_rw(struct Drbd_Conf *mdev, int rw)
+STATIC int bm_rw(struct drbd_conf *mdev, int rw) __must_hold(local)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	/* sector_t sector; */
@@ -701,37 +670,34 @@ STATIC int bm_rw(struct Drbd_Conf *mdev, int rw)
 	char ppb[10];
 	int err = 0;
 
-	MUST_BE_LOCKED();
+	WARN_ON(!bm_is_locked(b));
+
+	/* no spinlock here, the drbd_bm_lock should be enough! */
 
 	bm_words  = drbd_bm_words(mdev);
 	num_pages = (bm_words*sizeof(long) + PAGE_SIZE-1) >> PAGE_SHIFT;
 
-	/* OK, I manipulate the bitmap low level,
-	 * and I expect to be the exclusive user.
-	 * If not, I am really in a bad mood...
-	 * to catch such bugs early, make all people who want to access the
-	 * bitmap while I read/write it dereference a NULL pointer :->
-	 */
-	mdev->bitmap = NULL;
-
 	/* on disk bitmap is little endian */
 	if (rw == WRITE)
 		bm_cpu_to_lel(b);
 
 	now = jiffies;
 	atomic_set(&b->bm_async_io, num_pages);
-	__clear_bit(BM_MD_IO_ERROR,&b->bm_flags);
+	__clear_bit(BM_MD_IO_ERROR, &b->bm_flags);
 
 	/* let the layers below us try to merge these bios... */
 	for (i = 0; i < num_pages; i++)
-		bm_page_io_async(mdev,b,i,rw);
+		bm_page_io_async(mdev, b, i, rw);
 
 	drbd_blk_run_queue(bdev_get_queue(mdev->bc->md_bdev));
 	wait_event(b->bm_io_wait, atomic_read(&b->bm_async_io) == 0);
-	INFO("%s of bitmap took %lu jiffies\n",
-	     rw == READ ? "reading" : "writing", jiffies - now);
 
-	if (test_bit(BM_MD_IO_ERROR,&b->bm_flags)) {
+	MTRACE(TraceTypeMDIO, TraceLvlSummary,
+	       INFO("%s of bitmap took %lu jiffies\n",
+		    rw == READ ? "reading" : "writing", jiffies - now);
+	       );
+
+	if (test_bit(BM_MD_IO_ERROR, &b->bm_flags)) {
 		ALERT("we had at least one MD IO ERROR during bitmap IO\n");
 		drbd_chk_io_error(mdev, 1, TRUE);
 		drbd_io_error(mdev, TRUE);
@@ -743,7 +709,7 @@ STATIC int bm_rw(struct Drbd_Conf *mdev, int rw)
 		/* swap back endianness */
 		bm_lel_to_cpu(b);
 		/* flush bitmap to stable storage */
-		if (!test_bit(MD_NO_BARRIER,&mdev->flags))
+		if (!test_bit(MD_NO_BARRIER, &mdev->flags))
 			blkdev_issue_flush(mdev->bc->md_bdev, NULL);
 	} else /* rw == READ */ {
 		/* just read, if neccessary adjust endianness */
@@ -753,14 +719,8 @@ STATIC int bm_rw(struct Drbd_Conf *mdev, int rw)
 	}
 	now = b->bm_set;
 
-	/* ok, done,
-	 * now it is visible again
-	 */
-
-	mdev->bitmap = b;
-
 	INFO("%s (%lu bits) marked out-of-sync by on disk bit-map.\n",
-	     ppsize(ppb,now << (BM_BLOCK_SIZE_B-10)), now);
+	     ppsize(ppb, now << (BM_BLOCK_SIZE_B-10)), now);
 
 	return err;
 }
@@ -770,16 +730,17 @@ STATIC int bm_rw(struct Drbd_Conf *mdev, int rw)
  *
  * currently only called from "drbd_nl_disk_conf"
  */
-int drbd_bm_read(struct Drbd_Conf *mdev)
+int drbd_bm_read(struct drbd_conf *mdev) __must_hold(local)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	int err=0;
+	int err = 0;
 
 	if (b->bm) {
-	    err = bm_rw(mdev, READ);
+		/* bitmap size > 0 */
+		err = bm_rw(mdev, READ);
 
-	    if (err == 0)
-		b->bm[b->bm_words] = DRBD_MAGIC;
+		if (err == 0)
+			b->bm[b->bm_words] = DRBD_MAGIC;
 	}
 
 	return err;
@@ -790,7 +751,7 @@ int drbd_bm_read(struct Drbd_Conf *mdev)
  *
  * called at various occasions.
  */
-int drbd_bm_write(struct Drbd_Conf *mdev)
+int drbd_bm_write(struct drbd_conf *mdev) __must_hold(local)
 {
 	return bm_rw(mdev, WRITE);
 }
@@ -802,53 +763,50 @@ int drbd_bm_write(struct Drbd_Conf *mdev)
  * @enr: The _sector_ offset from the start of the bitmap.
  *
  */
-int drbd_bm_write_sect(struct Drbd_Conf *mdev,unsigned long enr)
+int drbd_bm_write_sect(struct drbd_conf *mdev, unsigned long enr) __must_hold(local)
 {
-	sector_t on_disk_sector = enr + mdev->bc->md.md_offset + mdev->bc->md.bm_offset;
-	int bm_words, num_words, offset, err  = 0;
-
-	// MUST_BE_LOCKED(); not neccessarily global...
+	sector_t on_disk_sector = enr + mdev->bc->md.md_offset
+				      + mdev->bc->md.bm_offset;
+	int bm_words, num_words, offset;
+	int err = 0;
 
 	down(&mdev->md_io_mutex);
 	bm_words  = drbd_bm_words(mdev);
-	offset    = S2W(enr);	// word offset into bitmap
+	offset    = S2W(enr);	/* word offset into bitmap */
 	num_words = min(S2W(1), bm_words - offset);
 #if DUMP_MD >= 3
 	INFO("write_sect: sector=%lu offset=%u num_words=%u\n",
 			enr, offset, num_words);
 #endif
-	if (num_words < S2W(1)) {
-		memset(page_address(mdev->md_io_page),0,MD_HARDSECT);
-	}
-	drbd_bm_get_lel( mdev, offset, num_words,
-			 page_address(mdev->md_io_page) );
-	if (!drbd_md_sync_page_io(mdev,mdev->bc,on_disk_sector,WRITE)) {
+	if (num_words < S2W(1))
+		memset(page_address(mdev->md_io_page), 0, MD_HARDSECT);
+	drbd_bm_get_lel(mdev, offset, num_words,
+			page_address(mdev->md_io_page));
+	if (!drbd_md_sync_page_io(mdev, mdev->bc, on_disk_sector, WRITE)) {
 		int i;
 		err = -EIO;
-		ERR( "IO ERROR writing bitmap sector %lu "
-		     "(meta-disk sector %llus)\n",
-		     enr, (unsigned long long)on_disk_sector );
+		ERR("IO ERROR writing bitmap sector %lu "
+		    "(meta-disk sector %llus)\n",
+		    enr, (unsigned long long)on_disk_sector);
 		drbd_chk_io_error(mdev, 1, TRUE);
 		drbd_io_error(mdev, TRUE);
 		for (i = 0; i < AL_EXT_PER_BM_SECT; i++)
-			drbd_bm_ALe_set_all(mdev,enr*AL_EXT_PER_BM_SECT+i);
+			drbd_bm_ALe_set_all(mdev, enr*AL_EXT_PER_BM_SECT+i);
 	}
 	mdev->bm_writ_cnt++;
 	up(&mdev->md_io_mutex);
 	return err;
 }
 
-void drbd_bm_reset_find(drbd_dev *mdev)
+void drbd_bm_reset_find(struct drbd_conf *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 
 	ERR_IF(!b) return;
 
-	MUST_BE_LOCKED();
-
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
 	b->bm_fo = 0;
 	spin_unlock_irq(&b->bm_lock);
 
@@ -859,7 +817,7 @@ void drbd_bm_reset_find(drbd_dev *mdev)
  * should not make much difference anyways, but ...
  * this returns a bit number, NOT a sector!
  */
-unsigned long drbd_bm_find_next(drbd_dev *mdev)
+unsigned long drbd_bm_find_next(struct drbd_conf *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long i = -1UL;
@@ -867,14 +825,14 @@ unsigned long drbd_bm_find_next(drbd_dev *mdev)
 	ERR_IF(!b) return i;
 	ERR_IF(!b->bm) return i;
 
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
-	if (b->bm_fo < b->bm_bits) {
-		i = find_next_bit(b->bm,b->bm_bits,b->bm_fo);
-	} else if (b->bm_fo > b->bm_bits) {
-		ERR("bm_fo=%lu bm_bits=%lu\n",b->bm_fo, b->bm_bits);
-	}
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
+	if (b->bm_fo < b->bm_bits)
+		i = find_next_bit(b->bm, b->bm_bits, b->bm_fo);
+	else if (b->bm_fo > b->bm_bits)
+		ERR("bm_fo=%lu bm_bits=%lu\n", b->bm_fo, b->bm_bits);
+
 	if (i >= b->bm_bits) {
 		i = -1UL;
 		/* leave b->bm_fo unchanged. */
@@ -885,13 +843,11 @@ unsigned long drbd_bm_find_next(drbd_dev *mdev)
 	return i;
 }
 
-void drbd_bm_set_find(drbd_dev *mdev, unsigned long i)
+void drbd_bm_set_find(struct drbd_conf *mdev, unsigned long i)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
 
 	b->bm_fo = min_t(unsigned long, i, b->bm_bits);
 
@@ -899,16 +855,18 @@ void drbd_bm_set_find(drbd_dev *mdev, unsigned long i)
 }
 
 
-int drbd_bm_rs_done(drbd_dev *mdev)
+int drbd_bm_rs_done(struct drbd_conf *mdev)
 {
-	return (mdev->bitmap->bm_fo >= mdev->bitmap->bm_bits);
+	D_ASSERT(mdev->bitmap);
+	return mdev->bitmap->bm_fo >= mdev->bitmap->bm_bits;
 }
 
 /* returns number of bits actually changed.
  * for val != 0, we change 0 -> 1, return code positiv
  * for val == 0, we change 1 -> 0, return code negative
  * wants bitnr, not sector */
-static int bm_change_bits_to(drbd_dev *mdev, const unsigned long s, const unsigned long e, int val)
+static int bm_change_bits_to(struct drbd_conf *mdev, const unsigned long s,
+	const unsigned long e, int val)
 {
 	unsigned long flags;
 	struct drbd_bitmap *b = mdev->bitmap;
@@ -917,12 +875,12 @@ static int bm_change_bits_to(drbd_dev *mdev, const unsigned long s, const unsign
 	ERR_IF(!b) return 1;
 	ERR_IF(!b->bm) return 1;
 
-	spin_lock_irqsave(&b->bm_lock,flags);
-	BM_PARANOIA_CHECK();
-	MUST_NOT_BE_LOCKED();
-	for (bitnr = s; bitnr <=e; bitnr++) {
+	spin_lock_irqsave(&b->bm_lock, flags);
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
+	for (bitnr = s; bitnr <= e; bitnr++) {
 		ERR_IF (bitnr >= b->bm_bits) {
-			ERR("bitnr=%lu bm_bits=%lu\n",bitnr, b->bm_bits);
+			ERR("bitnr=%lu bm_bits=%lu\n", bitnr, b->bm_bits);
 		} else {
 			if (val)
 				c += (0 == __test_and_set_bit(bitnr, b->bm));
@@ -931,18 +889,18 @@ static int bm_change_bits_to(drbd_dev *mdev, const unsigned long s, const unsign
 		}
 	}
 	b->bm_set += c;
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irqrestore(&b->bm_lock, flags);
 	return c;
 }
 
 /* returns number of bits changed 0 -> 1 */
-int drbd_bm_set_bits(drbd_dev *mdev, const unsigned long s, const unsigned long e)
+int drbd_bm_set_bits(struct drbd_conf *mdev, const unsigned long s, const unsigned long e)
 {
 	return bm_change_bits_to(mdev, s, e, 1);
 }
 
 /* returns number of bits changed 1 -> 0 */
-int drbd_bm_clear_bits(drbd_dev *mdev, const unsigned long s, const unsigned long e)
+int drbd_bm_clear_bits(struct drbd_conf *mdev, const unsigned long s, const unsigned long e)
 {
 	return -bm_change_bits_to(mdev, s, e, 0);
 }
@@ -954,7 +912,7 @@ int drbd_bm_clear_bits(drbd_dev *mdev, const unsigned long s, const unsigned lon
  *  0 ... bit not set
  * -1 ... first out of bounds access, stop testing for bits!
  */
-int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
+int drbd_bm_test_bit(struct drbd_conf *mdev, const unsigned long bitnr)
 {
 	unsigned long flags;
 	struct drbd_bitmap *b = mdev->bitmap;
@@ -963,13 +921,14 @@ int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
 	ERR_IF(!b->bm) return 0;
 
 	spin_lock_irqsave(&b->bm_lock, flags);
-	BM_PARANOIA_CHECK();
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
 	if (bitnr < b->bm_bits) {
 		i = test_bit(bitnr, b->bm) ? 1 : 0;
 	} else if (bitnr == b->bm_bits) {
 		i = -1;
-	} else /* (bitnr > b->bm_bits) */ {
-		ERR("bitnr=%lu > bm_bits=%lu\n",bitnr, b->bm_bits);
+	} else { /* (bitnr > b->bm_bits) */
+		ERR("bitnr=%lu > bm_bits=%lu\n", bitnr, b->bm_bits);
 		i = 0;
 	}
 
@@ -978,7 +937,7 @@ int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
 }
 
 /* returns number of bits set */
-int drbd_bm_count_bits(drbd_dev *mdev, const unsigned long s, const unsigned long e)
+int drbd_bm_count_bits(struct drbd_conf *mdev, const unsigned long s, const unsigned long e)
 {
 	unsigned long flags;
 	struct drbd_bitmap *b = mdev->bitmap;
@@ -992,17 +951,15 @@ int drbd_bm_count_bits(drbd_dev *mdev, const unsigned long s, const unsigned lon
 	ERR_IF(!b) return 1;
 	ERR_IF(!b->bm) return 1;
 
-	spin_lock_irqsave(&b->bm_lock,flags);
-	BM_PARANOIA_CHECK();
-	MUST_NOT_BE_LOCKED();
-	for (bitnr = s; bitnr <=e; bitnr++) {
+	spin_lock_irqsave(&b->bm_lock, flags);
+	for (bitnr = s; bitnr <= e; bitnr++) {
 		ERR_IF (bitnr >= b->bm_bits) {
-			ERR("bitnr=%lu bm_bits=%lu\n",bitnr, b->bm_bits);
+			ERR("bitnr=%lu bm_bits=%lu\n", bitnr, b->bm_bits);
 		} else {
 			c += (0 != test_bit(bitnr, b->bm));
 		}
 	}
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irqrestore(&b->bm_lock, flags);
 	return c;
 }
 
@@ -1021,7 +978,7 @@ int drbd_bm_count_bits(drbd_dev *mdev, const unsigned long s, const unsigned lon
  * reference count of some bitmap extent element from some lru instead...
  *
  */
-int drbd_bm_e_weight(drbd_dev *mdev, unsigned long enr)
+int drbd_bm_e_weight(struct drbd_conf *mdev, unsigned long enr)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	int count, s, e;
@@ -1029,20 +986,22 @@ int drbd_bm_e_weight(drbd_dev *mdev, unsigned long enr)
 
 	ERR_IF(!b) return 0;
 	ERR_IF(!b->bm) return 0;
-	spin_lock_irqsave(&b->bm_lock,flags);
-	BM_PARANOIA_CHECK();
+	spin_lock_irqsave(&b->bm_lock, flags);
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
 
 	s = S2W(enr);
-	e = min((size_t)S2W(enr+1),b->bm_words);
+	e = min((size_t)S2W(enr+1), b->bm_words);
 	count = 0;
 	if (s < b->bm_words) {
-		const unsigned long* w = b->bm+s;
+		const unsigned long *w = b->bm+s;
 		int n = e-s;
-		while (n--) count += hweight_long(*w++);
+		while (n--)
+			count += hweight_long(*w++);
 	} else {
 		ERR("start offset (%d) too large in drbd_bm_e_weight\n", s);
 	}
-	spin_unlock_irqrestore(&b->bm_lock,flags);
+	spin_unlock_irqrestore(&b->bm_lock, flags);
 #if DUMP_MD >= 3
 	INFO("enr=%lu weight=%d e=%d s=%d\n", enr, count, e, s);
 #endif
@@ -1050,7 +1009,7 @@ int drbd_bm_e_weight(drbd_dev *mdev, unsigned long enr)
 }
 
 /* set all bits covered by the AL-extent al_enr */
-unsigned long drbd_bm_ALe_set_all(drbd_dev *mdev, unsigned long al_enr)
+unsigned long drbd_bm_ALe_set_all(struct drbd_conf *mdev, unsigned long al_enr)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long weight;
@@ -1058,26 +1017,24 @@ unsigned long drbd_bm_ALe_set_all(drbd_dev *mdev, unsigned long al_enr)
 	ERR_IF(!b) return 0;
 	ERR_IF(!b->bm) return 0;
 
-	MUST_BE_LOCKED();
-
-	D_BUG_ON(irqs_disabled());
 	spin_lock_irq(&b->bm_lock);
-	BM_PARANOIA_CHECK();
+	if (bm_is_locked(b))
+		bm_print_lock_info(mdev);
 	weight = b->bm_set;
 
 	s = al_enr * BM_WORDS_PER_AL_EXT;
 	e = min_t(size_t, s + BM_WORDS_PER_AL_EXT, b->bm_words);
 	count = 0;
 	if (s < b->bm_words) {
-		const unsigned long* w = b->bm+s;
+		const unsigned long *w = b->bm+s;
 		int n = e-s;
-		while (n--) count += hweight_long(*w++);
+		while (n--)
+			count += hweight_long(*w++);
 		n = e-s;
-		memset(b->bm+s,-1,n*sizeof(long));
+		memset(b->bm+s, -1, n*sizeof(long));
 		b->bm_set += n*BITS_PER_LONG - count;
-		if (e == b->bm_words) {
+		if (e == b->bm_words)
 			b->bm_set -= bm_clear_surplus(b);
-		}
 	} else {
 		ERR("start offset (%d) too large in drbd_bm_ALe_set_all\n", s);
 	}
