@@ -34,7 +34,7 @@
 #include <linux/wait.h>
 #include <linux/mm.h>
 #include <linux/drbd_config.h>
-#include <linux/mm_inline.h> // for the page_count macro on RH/Fedora
+#include <linux/mm_inline.h>
 #include <linux/slab.h>
 #include <linux/random.h>
 
@@ -62,7 +62,7 @@
 /* used for synchronous meta data and bitmap IO
  * submitted by drbd_md_sync_page_io()
  */
-BIO_ENDIO_FN(drbd_md_io_complete)
+BIO_ENDIO_TYPE drbd_md_io_complete BIO_ENDIO_ARGS(struct bio *bio, int error)
 {
 	struct drbd_md_io *md_io;
 
@@ -83,25 +83,22 @@ BIO_ENDIO_FN(drbd_md_io_complete)
 /* reads on behalf of the partner,
  * "submitted" by the receiver
  */
-BIO_ENDIO_FN(drbd_endio_read_sec)
+BIO_ENDIO_TYPE drbd_endio_read_sec BIO_ENDIO_ARGS(struct bio *bio, int error) __releases(local)
 {
-	unsigned long flags=0;
-	struct Tl_epoch_entry *e=NULL;
-	struct Drbd_Conf* mdev;
+	unsigned long flags = 0;
+	struct Tl_epoch_entry *e = NULL;
+	struct drbd_conf *mdev;
 	int uptodate = bio_flagged(bio, BIO_UPTODATE);
 
 	e = bio->bi_private;
 	mdev = e->mdev;
 
-	/* We are called each time a part of the bio is finished, but
-	 * we are only interested when the whole bio is finished, therefore
-	 * return as long as bio->bio_size is positive.  */
 	BIO_ENDIO_FN_START;
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
 		 * but do not return any error?!
-		 * do we want to WARN() on this? */
+		 * do we want to drbd_WARN() on this? */
 		error = -EIO;
 	}
 
@@ -109,19 +106,20 @@ BIO_ENDIO_FN(drbd_endio_read_sec)
 
 	dump_internal_bio("Sec", mdev, bio, 1);
 
-	spin_lock_irqsave(&mdev->req_lock,flags);
+	spin_lock_irqsave(&mdev->req_lock, flags);
 	mdev->read_cnt += e->size >> 9;
 	list_del(&e->w.list);
-	if(list_empty(&mdev->read_ee)) wake_up(&mdev->ee_wait);
-	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	if (list_empty(&mdev->read_ee))
+		wake_up(&mdev->ee_wait);
+	spin_unlock_irqrestore(&mdev->req_lock, flags);
 
-	drbd_chk_io_error(mdev,error,FALSE);
-	drbd_queue_work(&mdev->data.work,&e->w);
+	drbd_chk_io_error(mdev, error, FALSE);
+	drbd_queue_work(&mdev->data.work, &e->w);
 	dec_local(mdev);
 
-	MTRACE(TraceTypeEE,TraceLvlAll,
+	MTRACE(TraceTypeEE, TraceLvlAll,
 	       INFO("Moved EE (READ) to worker sec=%llus size=%u ee=%p\n",
-		    (unsigned long long)e->sector,e->size,e);
+		    (unsigned long long)e->sector, e->size, e);
 	       );
 	BIO_ENDIO_FN_RETURN;
 }
@@ -129,11 +127,11 @@ BIO_ENDIO_FN(drbd_endio_read_sec)
 /* writes on behalf of the partner, or resync writes,
  * "submitted" by the receiver.
  */
-BIO_ENDIO_FN(drbd_endio_write_sec)
+BIO_ENDIO_TYPE drbd_endio_write_sec BIO_ENDIO_ARGS(struct bio *bio, int error) __releases(local)
 {
-	unsigned long flags=0;
-	struct Tl_epoch_entry *e=NULL;
-	drbd_dev *mdev;
+	unsigned long flags = 0;
+	struct Tl_epoch_entry *e = NULL;
+	struct drbd_conf *mdev;
 	sector_t e_sector;
 	int do_wake;
 	int is_syncer_req;
@@ -148,7 +146,7 @@ BIO_ENDIO_FN(drbd_endio_write_sec)
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
 		 * but do not return any error?!
-		 * do we want to WARN() on this? */
+		 * do we want to drbd_WARN() on this? */
 		error = -EIO;
 	}
 
@@ -156,7 +154,7 @@ BIO_ENDIO_FN(drbd_endio_write_sec)
 
 	dump_internal_bio("Sec", mdev, bio, 1);
 
-	spin_lock_irqsave(&mdev->req_lock,flags);
+	spin_lock_irqsave(&mdev->req_lock, flags);
 	mdev->writ_cnt += e->size >> 9;
 	is_syncer_req = is_syncer_block_id(e->block_id);
 
@@ -168,11 +166,11 @@ BIO_ENDIO_FN(drbd_endio_write_sec)
 	do_al_complete_io = e->flags & EE_CALL_AL_COMPLETE_IO;
 
 	list_del(&e->w.list); /* has been on active_ee or sync_ee */
-	list_add_tail(&e->w.list,&mdev->done_ee);
+	list_add_tail(&e->w.list, &mdev->done_ee);
 
-	MTRACE(TraceTypeEE,TraceLvlAll,
+	MTRACE(TraceTypeEE, TraceLvlAll,
 	       INFO("Moved EE (WRITE) to done_ee sec=%llus size=%u ee=%p\n",
-		    (unsigned long long)e->sector,e->size,e);
+		    (unsigned long long)e->sector, e->size, e);
 	       );
 
 	/* No hlist_del_init(&e->colision) here, we did not send the Ack yet,
@@ -180,20 +178,25 @@ BIO_ENDIO_FN(drbd_endio_write_sec)
 	 * done from "drbd_process_done_ee" within the appropriate w.cb
 	 * (e_end_block/e_end_resync_block) or from _drbd_clear_done_ee */
 
-	if(!is_syncer_req) mdev->epoch_size++;
+	if (!is_syncer_req)
+		mdev->epoch_size++;
 
 	do_wake = is_syncer_req
 		? list_empty(&mdev->sync_ee)
 		: list_empty(&mdev->active_ee);
 
-	if (error) __drbd_chk_io_error(mdev,FALSE);
-	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	if (error)
+		__drbd_chk_io_error(mdev, FALSE);
+	spin_unlock_irqrestore(&mdev->req_lock, flags);
 
-	if (is_syncer_req) drbd_rs_complete_io(mdev,e_sector);
+	if (is_syncer_req)
+		drbd_rs_complete_io(mdev, e_sector);
 
-	if (do_wake) wake_up(&mdev->ee_wait);
+	if (do_wake)
+		wake_up(&mdev->ee_wait);
 
-	if (do_al_complete_io) drbd_al_complete_io(mdev,e_sector);
+	if (do_al_complete_io)
+		drbd_al_complete_io(mdev, e_sector);
 
 	wake_asender(mdev);
 	dec_local(mdev);
@@ -203,21 +206,20 @@ BIO_ENDIO_FN(drbd_endio_write_sec)
 
 /* read, readA or write requests on Primary comming from drbd_make_request
  */
-BIO_ENDIO_FN(drbd_endio_pri)
+BIO_ENDIO_TYPE drbd_endio_pri BIO_ENDIO_ARGS(struct bio *bio, int error)
 {
 	unsigned long flags;
-	drbd_request_t *req=bio->bi_private;
-	drbd_dev *mdev = req->mdev;
-	drbd_req_event_t what;
+	struct drbd_request *req = bio->bi_private;
+	struct drbd_conf *mdev = req->mdev;
+	enum drbd_req_event what;
 	int uptodate = bio_flagged(bio, BIO_UPTODATE);
 
 	BIO_ENDIO_FN_START;
-
 	if (!error && !uptodate) {
 		/* strange behaviour of some lower level drivers...
 		 * fail the request by clearing the uptodate flag,
 		 * but do not return any error?!
-		 * do we want to WARN() on this? */
+		 * do we want to drbd_WARN() on this? */
 		error = -EIO;
 	}
 
@@ -226,18 +228,18 @@ BIO_ENDIO_FN(drbd_endio_pri)
 	/* to avoid recursion in _req_mod */
 	what = error
 	       ? (bio_data_dir(bio) == WRITE)
-	         ? write_completed_with_error
-	         : read_completed_with_error
+		 ? write_completed_with_error
+		 : read_completed_with_error
 	       : completed_ok;
-	spin_lock_irqsave(&mdev->req_lock,flags);
+	spin_lock_irqsave(&mdev->req_lock, flags);
 	_req_mod(req, what, error);
-	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	spin_unlock_irqrestore(&mdev->req_lock, flags);
 	BIO_ENDIO_FN_RETURN;
 }
 
-int w_io_error(drbd_dev* mdev, struct drbd_work* w,int cancel)
+int w_io_error(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	drbd_request_t *req = (drbd_request_t*)w;
+	struct drbd_request *req = (struct drbd_request *)w;
 	int ok;
 
 	/* FIXME send a "set_out_of_sync" packet to the peer
@@ -246,98 +248,99 @@ int w_io_error(drbd_dev* mdev, struct drbd_work* w,int cancel)
 	 * a "we are diskless" param packet anyways, and the peer
 	 * will then set the FullSync bit in the meta data ...
 	 */
-	// NOTE: mdev->bc can be NULL by the time we get here!
-	//D_ASSERT(mdev->bc->dc.on_io_error != PassOn);
+	/* NOTE: mdev->bc can be NULL by the time we get here! */
+	/* D_ASSERT(mdev->bc->dc.on_io_error != PassOn); */
 
 	/* the only way this callback is scheduled is from _req_may_be_done,
 	 * when it is done and had a local write error, see comments there */
 	drbd_req_free(req);
 
 	ok = drbd_io_error(mdev, FALSE);
-	if(unlikely(!ok)) ERR("Sending in w_io_error() failed\n");
+	if (unlikely(!ok))
+		ERR("Sending in w_io_error() failed\n");
 	return ok;
 }
 
-int w_read_retry_remote(drbd_dev* mdev, struct drbd_work* w,int cancel)
+int w_read_retry_remote(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	drbd_request_t *req = (drbd_request_t*)w;
+	struct drbd_request *req = (struct drbd_request *)w;
 
 	/* FIXME this is ugly. we should not detach for read io-error,
 	 * but try to WRITE the DataReply to the failed location,
 	 * to give the disk the chance to relocate that block */
-	drbd_io_error(mdev,FALSE); /* tries to schedule a detach and notifies peer */
+	drbd_io_error(mdev, FALSE); /* tries to schedule a detach and notifies peer */
 
 	spin_lock_irq(&mdev->req_lock);
-	if ( cancel ||
-	     mdev->state.conn < Connected ||
-	     mdev->state.pdsk <= Inconsistent ) {
+	if (cancel ||
+	    mdev->state.conn < Connected ||
+	    mdev->state.pdsk <= Inconsistent) {
 		_req_mod(req, send_canceled, 0); /* FIXME freeze? ... */
 		spin_unlock_irq(&mdev->req_lock);
-		drbd_khelper(mdev,"pri-on-incon-degr"); /* FIXME REALLY? */
 		ALERT("WE ARE LOST. Local IO failure, no peer.\n");
 		return 1;
 	}
 	spin_unlock_irq(&mdev->req_lock);
 
-	return w_send_read_req(mdev,w,0);
+	return w_send_read_req(mdev, w, 0);
 }
 
-int w_resync_inactive(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_resync_inactive(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
 	ERR_IF(cancel) return 1;
 	ERR("resync inactive, but callback triggered??\n");
-	return 1; // Simply ignore this!
+	return 1; /* Simply ignore this! */
 }
 
 void resync_timer_fn(unsigned long data)
 {
 	unsigned long flags;
-	drbd_dev* mdev = (drbd_dev*) data;
+	struct drbd_conf *mdev = (struct drbd_conf *) data;
 	int queue;
 
-	spin_lock_irqsave(&mdev->req_lock,flags);
+	spin_lock_irqsave(&mdev->req_lock, flags);
 
-	if(likely(!test_and_clear_bit(STOP_SYNC_TIMER,&mdev->flags))) {
-		queue=1;
+	if (likely(!test_and_clear_bit(STOP_SYNC_TIMER, &mdev->flags))) {
+		queue = 1;
 		mdev->resync_work.cb = w_make_resync_request;
 	} else {
-		queue=0;
+		queue = 0;
 		mdev->resync_work.cb = w_resync_inactive;
 	}
 
-	spin_unlock_irqrestore(&mdev->req_lock,flags);
+	spin_unlock_irqrestore(&mdev->req_lock, flags);
 
 	/* harmless race: list_empty outside data.work.q_lock */
-	if(list_empty(&mdev->resync_work.list) && queue) {
-		drbd_queue_work(&mdev->data.work,&mdev->resync_work);
-	}
+	if (list_empty(&mdev->resync_work.list) && queue)
+		drbd_queue_work(&mdev->data.work, &mdev->resync_work);
 }
 
 #define SLEEP_TIME (HZ/10)
 
-int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
+int w_make_resync_request(struct drbd_conf *mdev,
+		struct drbd_work *w, int cancel)
 {
 	unsigned long bit;
 	sector_t sector;
 	const sector_t capacity = drbd_get_capacity(mdev->this_bdev);
 	int max_segment_size = mdev->rq_queue->max_segment_size;
-	int number,i,size;
+	int number, i, size;
 	int align;
 
 	PARANOIA_BUG_ON(w != &mdev->resync_work);
 
-	if(unlikely(cancel)) return 1;
+	if (unlikely(cancel))
+		return 1;
 
-	if(unlikely(mdev->state.conn < Connected)) {
+	if (unlikely(mdev->state.conn < Connected)) {
 		ERR("Confused in w_make_resync_request()! cstate < Connected");
 		return 0;
 	}
 
-	if (mdev->state.conn != SyncTarget) {
-		ERR("%s in w_make_resync_request\n", conns_to_name(mdev->state.conn));
-	}
+	if (mdev->state.conn != SyncTarget)
+		ERR("%s in w_make_resync_request\n",
+			conns_to_name(mdev->state.conn));
 
-	if(!inc_local(mdev)) {
+	if (!inc_local(mdev)) {
 		/* Since we only need to access mdev->rsync a
 		   inc_local_if_state(mdev,Failed) would be sufficient, but
 		   to continue resync with a broken disk makes no sense at
@@ -350,14 +353,12 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 
 	number = SLEEP_TIME*mdev->sync_conf.rate / ((BM_BLOCK_SIZE/1024)*HZ);
 
-	if (atomic_read(&mdev->rs_pending_cnt)>number) {
+	if (atomic_read(&mdev->rs_pending_cnt) > number)
 		goto requeue;
-	}
 	number -= atomic_read(&mdev->rs_pending_cnt);
 
-	for(i=0;i<number;i++) {
-
-	next_sector:
+	for (i = 0; i < number; i++) {
+next_sector:
 		size = BM_BLOCK_SIZE;
 		/* as of now, we are the only user of drbd_bm_find_next */
 		bit  = drbd_bm_find_next(mdev);
@@ -374,13 +375,12 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 		sector = BM_BIT_TO_SECT(bit);
 
 		if (drbd_try_rs_begin_io(mdev, sector)) {
-			drbd_bm_set_find(mdev,bit);
+			drbd_bm_set_find(mdev, bit);
 			goto requeue;
 		}
 
-		if (unlikely(drbd_bm_test_bit(mdev,bit) == 0 )) {
-		      //INFO("Block got synced while in drbd_rs_begin_io()\n");
-			drbd_rs_complete_io(mdev,sector);
+		if (unlikely(drbd_bm_test_bit(mdev, bit) == 0)) {
+			drbd_rs_complete_io(mdev, sector);
 			goto next_sector;
 		}
 
@@ -397,41 +397,43 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 		 * "indirect" devices, this is dead code, since
 		 * q->max_segment_size will be PAGE_SIZE.
 		 */
-		align=1;
+		align = 1;
 		for (;;) {
 			if (size + BM_BLOCK_SIZE > max_segment_size)
 				break;
 
-			// Be always aligned
-			if (sector & ((1<<(align+3))-1) )
+			/* Be always aligned */
+			if (sector & ((1<<(align+3))-1))
 				break;
 
-			// do not cross extent boundaries
-			if (( (bit+1) & BM_BLOCKS_PER_BM_EXT_MASK ) == 0)
+			/* do not cross extent boundaries */
+			if (((bit+1) & BM_BLOCKS_PER_BM_EXT_MASK) == 0)
 				break;
 			/* now, is it actually dirty, after all?
 			 * caution, drbd_bm_test_bit is tri-state for some
 			 * obscure reason; ( b == 0 ) would get the out-of-band
 			 * only accidentally right because of the "oddly sized"
 			 * adjustment below */
-			if ( drbd_bm_test_bit(mdev,bit+1) != 1 )
+			if (drbd_bm_test_bit(mdev, bit+1) != 1)
 				break;
 			bit++;
 			size += BM_BLOCK_SIZE;
-			if( (BM_BLOCK_SIZE<<align) <= size) align++;
+			if ((BM_BLOCK_SIZE << align) <= size)
+				align++;
 			i++;
 		}
 		/* if we merged some,
 		 * reset the offset to start the next drbd_bm_find_next from */
 		if (size > BM_BLOCK_SIZE)
-			drbd_bm_set_find(mdev,bit+1);
+			drbd_bm_set_find(mdev, bit+1);
 #endif
 
 		/* adjust very last sectors, in case we are oddly sized */
-		if (sector + (size>>9) > capacity) size = (capacity-sector)<<9;
+		if (sector + (size>>9) > capacity)
+			size = (capacity-sector)<<9;
 		inc_rs_pending(mdev);
-		if(!drbd_send_drequest(mdev,RSDataRequest,
-				       sector,size,ID_SYNCER)) {
+		if (!drbd_send_drequest(mdev, RSDataRequest,
+				       sector, size, ID_SYNCER)) {
 			ERR("drbd_send_drequest() failed, aborting...\n");
 			dec_rs_pending(mdev);
 			dec_local(mdev);
@@ -439,7 +441,7 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 		}
 	}
 
-	if(drbd_bm_rs_done(mdev)) {
+	if (drbd_bm_rs_done(mdev)) {
 		/* last syncer _request_ was sent,
 		 * but the RSDataReply not yet received.  sync will end (and
 		 * next sync group will resume), as soon as we receive the last
@@ -457,226 +459,239 @@ int w_make_resync_request(drbd_dev* mdev, struct drbd_work* w,int cancel)
 	return 1;
 }
 
-int w_resync_finished(drbd_dev *mdev, struct drbd_work *w, int cancel)
+STATIC int w_resync_finished(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
 	kfree(w);
 
-	drbd_bm_lock(mdev);
 	drbd_resync_finished(mdev);
-	drbd_bm_unlock(mdev);
 
 	return 1;
 }
 
-int drbd_resync_finished(drbd_dev* mdev)
+int drbd_resync_finished(struct drbd_conf *mdev)
 {
-	unsigned long db,dt,dbdt;
-	int dstate, pdstate;
+	unsigned long db, dt, dbdt;
+	union drbd_state_t os, ns;
 	struct drbd_work *w;
 
-	// Remove all elements from the resync LRU. Since future actions
-	// might set bits in the (main) bitmap, then the entries in the
-	// resync LRU would be wrong.
-	if(drbd_rs_del_all(mdev)) {
-		// In case this is not possible now, most probabely because
-		// there are RSDataReply Packets lingering on the worker's
-		// queue (or even the read operations for those packets
-		// is not finished by now).   Retry in 100ms.
+	/* Remove all elements from the resync LRU. Since future actions
+	 * might set bits in the (main) bitmap, then the entries in the
+	 * resync LRU would be wrong. */
+	if (drbd_rs_del_all(mdev)) {
+		/* In case this is not possible now, most probabely because
+		 * there are RSDataReply Packets lingering on the worker's
+		 * queue (or even the read operations for those packets
+		 * is not finished by now).   Retry in 100ms. */
 
 		drbd_kick_lo(mdev);
-		set_current_state(TASK_INTERRUPTIBLE);
+		__set_current_state(TASK_INTERRUPTIBLE);
 		schedule_timeout(HZ / 10);
 		w = kmalloc(sizeof(struct drbd_work), GFP_ATOMIC);
-		if(w) {
+		if (w) {
 			w->cb = w_resync_finished;
-			drbd_queue_work(&mdev->data.work,w);
+			drbd_queue_work(&mdev->data.work, w);
 			return 1;
 		}
 		ERR("Warn failed to drbd_rs_del_all() and to kmalloc(w).\n");
 	}
 
 	dt = (jiffies - mdev->rs_start - mdev->rs_paused) / HZ;
-	if (dt <= 0) dt=1;
+	if (dt <= 0)
+		dt = 1;
 	db = mdev->rs_total;
 	dbdt = Bit2KB(db/dt);
 	mdev->rs_paused /= HZ;
+
+	if (!inc_local(mdev))
+		goto out;
+
+	spin_lock_irq(&mdev->req_lock);
+	os = mdev->state;
+
+	/* This protects us against multiple calls (that can happen in the presence
+	   of application IO), and against connectivity loss just before we arrive here. */
+	if (os.conn <= Connected)
+		goto out_unlock;
+
+	ns = os;
+	ns.conn = Connected;
+
 	INFO("Resync done (total %lu sec; paused %lu sec; %lu K/sec)\n",
 	     dt + mdev->rs_paused, mdev->rs_paused, dbdt);
 
 	D_ASSERT((drbd_bm_total_weight(mdev)-mdev->rs_failed) == 0);
 
 	if (mdev->rs_failed) {
-		INFO("            %lu failed blocks\n",mdev->rs_failed);
+		INFO("            %lu failed blocks\n", mdev->rs_failed);
 
-		if (mdev->state.conn == SyncTarget ||
-		    mdev->state.conn == PausedSyncT) {
-			dstate = Inconsistent;
-			pdstate = UpToDate;
+		if (os.conn == SyncTarget || os.conn == PausedSyncT) {
+			ns.disk = Inconsistent;
+			ns.pdsk = UpToDate;
 		} else {
-			dstate = UpToDate;
-			pdstate = Inconsistent;
+			ns.disk = UpToDate;
+			ns.pdsk = Inconsistent;
 		}
 	} else {
-		dstate = pdstate = UpToDate;
+		ns.disk = UpToDate;
+		ns.pdsk = UpToDate;
 
-		if (mdev->state.conn == SyncTarget ||
-		    mdev->state.conn == PausedSyncT) {
-			if( mdev->p_uuid ) {
+		if (os.conn == SyncTarget || os.conn == PausedSyncT) {
+			if (mdev->p_uuid) {
 				int i;
-				for ( i=Bitmap ; i<=History_end ; i++ ) {
-					_drbd_uuid_set(mdev,i,mdev->p_uuid[i]);
-				}
-				drbd_uuid_set(mdev,Bitmap,mdev->bc->md.uuid[Current]);
-				_drbd_uuid_set(mdev,Current,mdev->p_uuid[Current]);
+				for (i = Bitmap ; i <= History_end ; i++)
+					_drbd_uuid_set(mdev, i, mdev->p_uuid[i]);
+				drbd_uuid_set(mdev, Bitmap, mdev->bc->md.uuid[Current]);
+				_drbd_uuid_set(mdev, Current, mdev->p_uuid[Current]);
 			} else {
 				ERR("mdev->p_uuid is NULL! BUG\n");
 			}
 		}
 
-		drbd_uuid_set_bm(mdev,0UL);
+		drbd_uuid_set_bm(mdev, 0UL);
 
-		if ( mdev->p_uuid ) {
-			// Now the two UUID sets are equal, update what we
-			// know of the peer.
+		if (mdev->p_uuid) {
+			/* Now the two UUID sets are equal, update what we
+			 * know of the peer. */
 			int i;
-			for ( i=Current ; i<=History_end ; i++ ) {
-				mdev->p_uuid[i]=mdev->bc->md.uuid[i];
-			}
+			for (i = Current ; i <= History_end ; i++)
+				mdev->p_uuid[i] = mdev->bc->md.uuid[i];
 		}
 	}
 
+	_drbd_set_state(mdev, ns, ChgStateVerbose, NULL);
+out_unlock:
+	spin_unlock_irq(&mdev->req_lock);
+	dec_local(mdev);
+out:
 	mdev->rs_total  = 0;
 	mdev->rs_failed = 0;
 	mdev->rs_paused = 0;
 
-	if (test_and_clear_bit(WRITE_BM_AFTER_RESYNC,&mdev->flags)) {
-		WARN("Writing the whole bitmap, due to failed kmalloc\n");
-		drbd_bm_write(mdev);
+	if (test_and_clear_bit(WRITE_BM_AFTER_RESYNC, &mdev->flags)) {
+		drbd_WARN("Writing the whole bitmap, due to failed kmalloc\n");
+		drbd_queue_bitmap_io(mdev, &drbd_bm_write, NULL, "write from resync_finished");
 	}
 
 	drbd_bm_recount_bits(mdev);
 
-	drbd_request_state(mdev,NS3(conn,Connected,
-				    disk,dstate,
-				    pdsk,pdstate));
-
-	drbd_md_sync(mdev);
-
 	return 1;
 }
 
 /**
  * w_e_end_data_req: Send the answer (DataReply) in response to a DataRequest.
  */
-int w_e_end_data_req(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_e_end_data_req(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	struct Tl_epoch_entry *e = (struct Tl_epoch_entry*)w;
+	struct Tl_epoch_entry *e = (struct Tl_epoch_entry *)w;
 	int ok;
 
-	if(unlikely(cancel)) {
-		drbd_free_ee(mdev,e);
+	if (unlikely(cancel)) {
+		drbd_free_ee(mdev, e);
 		dec_unacked(mdev);
 		return 1;
 	}
 
-	if(likely(drbd_bio_uptodate(e->private_bio))) {
-		ok=drbd_send_block(mdev, DataReply, e);
+	if (likely(drbd_bio_uptodate(e->private_bio))) {
+		ok = drbd_send_block(mdev, DataReply, e);
 	} else {
-		if (DRBD_ratelimit(5*HZ,5))
+		if (DRBD_ratelimit(5*HZ, 5))
 			ERR("Sending NegDReply. sector=%llus.\n",
 			    (unsigned long long)e->sector);
 
-		ok=drbd_send_ack(mdev,NegDReply,e);
+		ok = drbd_send_ack(mdev, NegDReply, e);
 
 		/* FIXME we should not detach for read io-errors, in particular
 		 * not now: when the peer asked us for our data, we are likely
 		 * the only remaining disk... */
-		drbd_io_error(mdev,FALSE);
+		drbd_io_error(mdev, FALSE);
 	}
 
 	dec_unacked(mdev);
 
 	spin_lock_irq(&mdev->req_lock);
-	if( drbd_bio_has_active_page(e->private_bio) ) {
+	if (drbd_bio_has_active_page(e->private_bio)) {
 		/* This might happen if sendpage() has not finished */
-		list_add_tail(&e->w.list,&mdev->net_ee);
+		list_add_tail(&e->w.list, &mdev->net_ee);
 	} else {
-		drbd_free_ee(mdev,e);
+		drbd_free_ee(mdev, e);
 	}
 	spin_unlock_irq(&mdev->req_lock);
 
-	if(unlikely(!ok)) ERR("drbd_send_block() failed\n");
+	if (unlikely(!ok))
+		ERR("drbd_send_block() failed\n");
 	return ok;
 }
 
 /**
  * w_e_end_rsdata_req: Send the answer (RSDataReply) to a RSDataRequest.
  */
-int w_e_end_rsdata_req(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_e_end_rsdata_req(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	struct Tl_epoch_entry *e = (struct Tl_epoch_entry*)w;
+	struct Tl_epoch_entry *e = (struct Tl_epoch_entry *)w;
 	int ok;
 
-	if(unlikely(cancel)) {
-		drbd_free_ee(mdev,e);
+	if (unlikely(cancel)) {
+		drbd_free_ee(mdev, e);
 		dec_unacked(mdev);
 		return 1;
 	}
 
-	if(inc_local_if_state(mdev,Failed)) {
-		drbd_rs_complete_io(mdev,e->sector);
+	if (inc_local_if_state(mdev, Failed)) {
+		drbd_rs_complete_io(mdev, e->sector);
 		dec_local(mdev);
 	}
 
-	if(likely(drbd_bio_uptodate(e->private_bio))) {
-		if (likely( mdev->state.pdsk >= Inconsistent )) {
+	if (likely(drbd_bio_uptodate(e->private_bio))) {
+		if (likely(mdev->state.pdsk >= Inconsistent)) {
 			inc_rs_pending(mdev);
-			ok=drbd_send_block(mdev, RSDataReply, e);
+			ok = drbd_send_block(mdev, RSDataReply, e);
 		} else {
-			if (DRBD_ratelimit(5*HZ,5))
-				ERR("Not sending RSDataReply, partner DISKLESS!\n");
-			ok=1;
+			if (DRBD_ratelimit(5*HZ, 5))
+				ERR("Not sending RSDataReply, "
+				    "partner DISKLESS!\n");
+			ok = 1;
 		}
 	} else {
-		if (DRBD_ratelimit(5*HZ,5))
+		if (DRBD_ratelimit(5*HZ, 5))
 			ERR("Sending NegRSDReply. sector %llus.\n",
 			    (unsigned long long)e->sector);
 
-		ok=drbd_send_ack(mdev,NegRSDReply,e);
+		ok = drbd_send_ack(mdev, NegRSDReply, e);
 
 		drbd_io_error(mdev, FALSE);
 
-		// update resync data with failure
+		/* update resync data with failure */
 		drbd_rs_failed_io(mdev, e->sector, e->size);
 	}
 
 	dec_unacked(mdev);
 
 	spin_lock_irq(&mdev->req_lock);
-	if( drbd_bio_has_active_page(e->private_bio) ) {
+	if (drbd_bio_has_active_page(e->private_bio)) {
 		/* This might happen if sendpage() has not finished */
-		list_add_tail(&e->w.list,&mdev->net_ee);
+		list_add_tail(&e->w.list, &mdev->net_ee);
 	} else {
-		drbd_free_ee(mdev,e);
+		drbd_free_ee(mdev, e);
 	}
 	spin_unlock_irq(&mdev->req_lock);
 
-	if(unlikely(!ok)) ERR("drbd_send_block() failed\n");
+	if (unlikely(!ok))
+		ERR("drbd_send_block() failed\n");
 	return ok;
 }
 
-int w_prev_work_done(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_prev_work_done(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	clear_bit(WORK_PENDING,&mdev->flags);
+	clear_bit(WORK_PENDING, &mdev->flags);
 	wake_up(&mdev->misc_wait);
 	return 1;
 }
 
-int w_send_barrier(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_send_barrier(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
 	struct drbd_barrier *b = (struct drbd_barrier *)w;
-	Drbd_Barrier_Packet *p = &mdev->data.sbuf.Barrier;
-	int ok=1;
+	struct Drbd_Barrier_Packet *p = &mdev->data.sbuf.Barrier;
+	int ok = 1;
 
 	/* really avoid racing with tl_clear.  w.cb may have been referenced
 	 * just before it was reassigned and requeued, so double check that.
@@ -696,24 +711,26 @@ int w_send_barrier(drbd_dev *mdev, struct drbd_work *w, int cancel)
 	/* inc_ap_pending was done where this was queued.
 	 * dec_ap_pending will be done in got_BarrierAck
 	 * or (on connection loss) in w_clear_epoch.  */
-	ok = _drbd_send_cmd(mdev,mdev->data.socket,Barrier,(Drbd_Header*)p,sizeof(*p),0);
+	ok = _drbd_send_cmd(mdev, mdev->data.socket, Barrier,
+				(struct Drbd_Header *)p, sizeof(*p), 0);
 	drbd_put_data_sock(mdev);
 
 	return ok;
 }
 
-int w_send_write_hint(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_send_write_hint(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	if (cancel) return 1;
-	return drbd_send_short_cmd(mdev,UnplugRemote);
+	if (cancel)
+		return 1;
+	return drbd_send_short_cmd(mdev, UnplugRemote);
 }
 
 /**
  * w_send_dblock: Send a mirrored write request.
  */
-int w_send_dblock(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_send_dblock(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	drbd_request_t *req = (drbd_request_t *)w;
+	struct drbd_request *req = (struct drbd_request *)w;
 	int ok;
 
 	if (unlikely(cancel)) {
@@ -721,8 +738,8 @@ int w_send_dblock(drbd_dev *mdev, struct drbd_work *w, int cancel)
 		return 1;
 	}
 
-	ok = drbd_send_dblock(mdev,req);
-	req_mod(req,ok ? handed_over_to_network : send_failed, 0);
+	ok = drbd_send_dblock(mdev, req);
+	req_mod(req, ok ? handed_over_to_network : send_failed, 0);
 
 	return ok;
 }
@@ -730,9 +747,9 @@ int w_send_dblock(drbd_dev *mdev, struct drbd_work *w, int cancel)
 /**
  * w_send_read_req: Send a read requests.
  */
-int w_send_read_req(drbd_dev *mdev, struct drbd_work *w, int cancel)
+int w_send_read_req(struct drbd_conf *mdev, struct drbd_work *w, int cancel)
 {
-	drbd_request_t *req = (drbd_request_t *)w;
+	struct drbd_request *req = (struct drbd_request *)w;
 	int ok;
 
 	if (unlikely(cancel)) {
@@ -743,57 +760,63 @@ int w_send_read_req(drbd_dev *mdev, struct drbd_work *w, int cancel)
 	ok = drbd_send_drequest(mdev, DataRequest, req->sector, req->size,
 				(unsigned long)req);
 
-	if(ok) {
-		req_mod(req, handed_over_to_network, 0);
-	} else {
-		/* ?? we set Timeout or BrokenPipe in drbd_send() */
+	if (!ok) {
+		/* ?? we set Timeout or BrokenPipe in drbd_send();
+		 * so this is probably redundant */
 		if (mdev->state.conn >= Connected)
-			drbd_force_state(mdev,NS(conn,NetworkFailure));
-		/* req_mod(req, send_failed); we should not fail it here,
-		 * we might have to "freeze" on disconnect.
-		 * handled by req_mod(req, connection_lost_while_pending);
-		 * in drbd_fail_pending_reads soon enough. */
+			drbd_force_state(mdev, NS(conn, NetworkFailure));
 	}
+	req_mod(req, ok ? handed_over_to_network : send_failed, 0);
 
 	return ok;
 }
 
-STATIC void drbd_global_lock(void)
+STATIC void drbd_global_lock(void) __acquires(drbd_global_lock)
 {
-	drbd_dev *mdev;
+	struct drbd_conf *mdev;
 	int i;
 
+	__acquire(drbd_global_lock);
 	local_irq_disable();
-	for (i=0; i < minor_count; i++) {
-		if(!(mdev = minor_to_mdev(i))) continue;
+	for (i = 0; i < minor_count; i++) {
+		mdev = minor_to_mdev(i);
+		if (!mdev)
+			continue;
 		spin_lock(&mdev->req_lock);
+		__release(&mdev->req_lock); /* annihilate the spin_lock's annotation here */
 	}
 }
 
-STATIC void drbd_global_unlock(void)
+STATIC void drbd_global_unlock(void) __releases(drbd_global_lock)
 {
-	drbd_dev *mdev;
+	struct drbd_conf *mdev;
 	int i;
 
-	for (i=0; i < minor_count; i++) {
-		if(!(mdev = minor_to_mdev(i))) continue;
+	for (i = 0; i < minor_count; i++) {
+		mdev = minor_to_mdev(i);
+		if (!mdev)
+			continue;
+		__acquire(&mdev->req_lock);
 		spin_unlock(&mdev->req_lock);
 	}
 	local_irq_enable();
+	__release(drbd_global_lock);
 }
 
-STATIC int _drbd_may_sync_now(drbd_dev *mdev)
+STATIC int _drbd_may_sync_now(struct drbd_conf *mdev)
 {
-	drbd_dev *odev = mdev;
+	struct drbd_conf *odev = mdev;
 
-	while(1) {
-		if( odev->sync_conf.after == -1 ) return 1;
+	while (1) {
+		if (odev->sync_conf.after == -1)
+			return 1;
 		odev = minor_to_mdev(odev->sync_conf.after);
 		ERR_IF(!odev) return 1;
-		if( (odev->state.conn >= SyncSource &&
+		if ((odev->state.conn >= SyncSource &&
 		     odev->state.conn <= PausedSyncT) ||
 		    odev->state.aftr_isp || odev->state.peer_isp ||
-		    odev->state.user_isp ) return 0;
+		    odev->state.user_isp)
+			return 0;
 	}
 }
 
@@ -803,18 +826,20 @@ STATIC int _drbd_may_sync_now(drbd_dev *mdev)
  * pause their resynchronisation.
  * Called from process context only ( ioctl and after_state_ch ).
  */
-STATIC int _drbd_pause_after(drbd_dev *mdev)
+STATIC int _drbd_pause_after(struct drbd_conf *mdev)
 {
-	drbd_dev *odev;
+	struct drbd_conf *odev;
 	int i, rv = 0;
 
-	for (i=0; i < minor_count; i++) {
-		if (!(odev = minor_to_mdev(i)) ||
-		    (odev->state.conn == StandAlone && odev->state.disk == Diskless) ) continue;
-		if (! _drbd_may_sync_now(odev)) {
-			rv |= ( _drbd_set_state(_NS(odev,aftr_isp,1),ChgStateHard)
-				!= SS_NothingToDo ) ;
-		}
+	for (i = 0; i < minor_count; i++) {
+		odev = minor_to_mdev(i);
+		if (!odev)
+			continue;
+		if (odev->state.conn == StandAlone && odev->state.disk == Diskless)
+			continue;
+		if (!_drbd_may_sync_now(odev))
+			rv |= (_drbd_set_state(_NS(odev, aftr_isp, 1), ChgStateHard, NULL)
+				!= SS_NothingToDo);
 	}
 
 	return rv;
@@ -826,38 +851,40 @@ STATIC int _drbd_pause_after(drbd_dev *mdev)
  * process, and causes them to resume.
  * Called from process context only ( ioctl and worker ).
  */
-STATIC int _drbd_resume_next(drbd_dev *mdev)
+STATIC int _drbd_resume_next(struct drbd_conf *mdev)
 {
-	drbd_dev *odev;
+	struct drbd_conf *odev;
 	int i, rv = 0;
 
-	for (i=0; i < minor_count; i++) {
-		if( !(odev = minor_to_mdev(i)) ) continue;
-		if ( odev->state.aftr_isp ) {
-			if (_drbd_may_sync_now(odev)) {
-				rv |= ( _drbd_set_state(_NS(odev,aftr_isp,0),ChgStateHard)
-					!= SS_NothingToDo ) ;
-			}
+	for (i = 0; i < minor_count; i++) {
+		odev = minor_to_mdev(i);
+		if (!odev)
+			continue;
+		if (odev->state.aftr_isp) {
+			if (_drbd_may_sync_now(odev))
+				rv |= (_drbd_set_state(_NS(odev, aftr_isp, 0),
+						       ChgStateHard, NULL)
+					!= SS_NothingToDo) ;
 		}
 	}
 	return rv;
 }
 
-void resume_next_sg(drbd_dev* mdev)
+void resume_next_sg(struct drbd_conf *mdev)
 {
 	drbd_global_lock();
 	_drbd_resume_next(mdev);
 	drbd_global_unlock();
 }
 
-void suspend_other_sg(drbd_dev* mdev)
+void suspend_other_sg(struct drbd_conf *mdev)
 {
 	drbd_global_lock();
 	_drbd_pause_after(mdev);
 	drbd_global_unlock();
 }
 
-void drbd_alter_sa(drbd_dev *mdev, int na)
+void drbd_alter_sa(struct drbd_conf *mdev, int na)
 {
 	int changes;
 
@@ -880,14 +907,14 @@ void drbd_alter_sa(drbd_dev *mdev, int na)
  * Note, this function might bring you directly into one of the
  * PausedSync* states.
  */
-void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
+void drbd_start_resync(struct drbd_conf *mdev, enum drbd_conns side)
 {
-	drbd_state_t ns;
-	int r=0;
+	union drbd_state_t ns;
+	int r = 0;
 
 	MTRACE(TraceTypeResync, TraceLvlSummary,
 	       INFO("Resync starting: side=%s\n",
-		    side==SyncTarget?"SyncTarget":"SyncSource");
+		    side == SyncTarget ? "SyncTarget" : "SyncSource");
 	    );
 
 	drbd_bm_recount_bits(mdev);
@@ -895,14 +922,21 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 	/* In case a previous resync run was aborted by an IO error... */
 	drbd_rs_cancel_all(mdev);
 
-	if(side == SyncTarget) {
+	drbd_state_lock(mdev);
+
+	if (!inc_local_if_state(mdev, Negotiating)) {
+		drbd_state_unlock(mdev);
+		return;
+	}
+
+	if (side == SyncTarget) {
 		drbd_bm_reset_find(mdev);
 	} else /* side == SyncSource */ {
 		u64 uuid;
 
 		get_random_bytes(&uuid, sizeof(u64));
 		drbd_uuid_set(mdev, Bitmap, uuid);
-		drbd_send_sync_uuid(mdev,uuid);
+		drbd_send_sync_uuid(mdev, uuid);
 
 		D_ASSERT(mdev->state.disk == UpToDate);
 	}
@@ -914,16 +948,18 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 
 	ns.conn = side;
 
-	if(side == SyncTarget) {
+	if (side == SyncTarget)
 		ns.disk = Inconsistent;
-	} else /* side == SyncSource */ {
+	else /* side == SyncSource */
 		ns.pdsk = Inconsistent;
-	}
 
-	r = _drbd_set_state(mdev,ns,ChgStateVerbose);
+	r = _drbd_set_state(mdev, ns, ChgStateVerbose, NULL);
 	ns = mdev->state;
 
-	if ( r == SS_Success ) {
+	if (ns.conn < Connected)
+		r = SS_UnknownError;
+
+	if (r == SS_Success) {
 		mdev->rs_total     =
 		mdev->rs_mark_left = drbd_bm_total_weight(mdev);
 		mdev->rs_failed    = 0;
@@ -933,21 +969,23 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 		_drbd_pause_after(mdev);
 	}
 	drbd_global_unlock();
+	drbd_state_unlock(mdev);
+	dec_local(mdev);
 
-	if ( r == SS_Success ) {
+	if (r == SS_Success) {
 		INFO("Began resync as %s (will sync %lu KB [%lu bits set]).\n",
 		     conns_to_name(ns.conn),
 		     (unsigned long) mdev->rs_total << (BM_BLOCK_SIZE_B-10),
 		     (unsigned long) mdev->rs_total);
 
-		if ( mdev->rs_total == 0 ) {
+		if (mdev->rs_total == 0) {
 			drbd_resync_finished(mdev);
 			return;
 		}
 
-		if( ns.conn == SyncTarget ) {
-			D_ASSERT(!test_bit(STOP_SYNC_TIMER,&mdev->flags));
-			mod_timer(&mdev->resync_timer,jiffies);
+		if (ns.conn == SyncTarget) {
+			D_ASSERT(!test_bit(STOP_SYNC_TIMER, &mdev->flags));
+			mod_timer(&mdev->resync_timer, jiffies);
 		}
 
 		drbd_md_sync(mdev);
@@ -956,24 +994,26 @@ void drbd_start_resync(drbd_dev *mdev, drbd_conns_t side)
 
 int drbd_worker(struct Drbd_thread *thi)
 {
-	drbd_dev *mdev = thi->mdev;
-	struct drbd_work *w = 0;
+	struct drbd_conf *mdev = thi->mdev;
+	struct drbd_work *w = NULL;
 	LIST_HEAD(work_list);
-	int intr=0,i;
+	int intr = 0, i;
 
 	sprintf(current->comm, "drbd%d_worker", mdev_to_minor(mdev));
 
 	while (get_t_state(thi) == Running) {
 
-		if(down_trylock(&mdev->data.work.s)) {
+		if (down_trylock(&mdev->data.work.s)) {
 			down(&mdev->data.mutex);
-			if(mdev->data.socket)drbd_tcp_flush(mdev->data.socket);
+			if (mdev->data.socket)
+				drbd_tcp_uncork(mdev->data.socket);
 			up(&mdev->data.mutex);
 
 			intr = down_interruptible(&mdev->data.work.s);
 
 			down(&mdev->data.mutex);
-			if(mdev->data.socket) drbd_tcp_cork(mdev->data.socket);
+			if (mdev->data.socket)
+				drbd_tcp_cork(mdev->data.socket);
 			up(&mdev->data.mutex);
 		}
 
@@ -985,12 +1025,13 @@ int drbd_worker(struct Drbd_thread *thi)
 			break;
 		}
 
-		if (get_t_state(thi) != Running) break;
+		if (get_t_state(thi) != Running)
+			break;
 		/* With this break, we have done a down() but not consumed
 		   the entry from the list. The cleanup code takes care of
 		   this...   */
 
-		w = 0;
+		w = NULL;
 		spin_lock_irq(&mdev->data.work.q_lock);
 		ERR_IF(list_empty(&mdev->data.work.q)) {
 			/* something terribly wrong in our logic.
@@ -1007,33 +1048,34 @@ int drbd_worker(struct Drbd_thread *thi)
 			spin_unlock_irq(&mdev->data.work.q_lock);
 			continue;
 		}
-		w = list_entry(mdev->data.work.q.next,struct drbd_work,list);
+		w = list_entry(mdev->data.work.q.next, struct drbd_work, list);
 		list_del_init(&w->list);
 		spin_unlock_irq(&mdev->data.work.q_lock);
 
-		if(!w->cb(mdev,w, mdev->state.conn < Connected )) {
-			//WARN("worker: a callback failed! \n");
+		if (!w->cb(mdev, w, mdev->state.conn < Connected)) {
+			/* drbd_WARN("worker: a callback failed! \n"); */
 			if (mdev->state.conn >= Connected)
-				drbd_force_state(mdev,NS(conn,NetworkFailure));
+				drbd_force_state(mdev,
+						NS(conn, NetworkFailure));
 		}
 	}
 
 	spin_lock_irq(&mdev->data.work.q_lock);
 	i = 0;
 	while (!list_empty(&mdev->data.work.q)) {
-		list_splice_init(&mdev->data.work.q,&work_list);
+		list_splice_init(&mdev->data.work.q, &work_list);
 		spin_unlock_irq(&mdev->data.work.q_lock);
 
-		while(!list_empty(&work_list)) {
-			w = list_entry(work_list.next, struct drbd_work,list);
+		while (!list_empty(&work_list)) {
+			w = list_entry(work_list.next, struct drbd_work, list);
 			list_del_init(&w->list);
-			w->cb(mdev,w,1);
+			w->cb(mdev, w, 1);
 			i++; /* dead debugging code */
 		}
 
 		spin_lock_irq(&mdev->data.work.q_lock);
 	}
-	sema_init(&mdev->data.work.s,0);
+	sema_init(&mdev->data.work.s, 0);
 	/* DANGEROUS race: if someone did queue his work within the spinlock,
 	 * but up() ed outside the spinlock, we could get an up() on the
 	 * semaphore without corresponding list entry.
@@ -1046,7 +1088,10 @@ int drbd_worker(struct Drbd_thread *thi)
 	 * from the times when the worker did not live as long as the
 	 * device.. */
 
-	D_ASSERT( mdev->state.disk == Diskless && mdev->state.conn == StandAlone );
+	D_ASSERT(mdev->state.disk == Diskless && mdev->state.conn == StandAlone);
+	/* _drbd_set_state only uses stop_nowait.
+	 * wait here for the Exiting receiver. */
+	drbd_thread_stop(&mdev->receiver);
 	drbd_mdev_cleanup(mdev);
 
 	INFO("worker terminated\n");
