@@ -26,6 +26,7 @@
 #ifndef DRBD_H
 #define DRBD_H
 #include <linux/drbd_config.h>
+#include <linux/connector.h>
 
 #include <asm/types.h>
 
@@ -37,7 +38,7 @@
 #include <sys/wait.h>
 #include <limits.h>
 
-/* Altough the Linux source code makes a difference between 
+/* Altough the Linux source code makes a difference between
    generic endiness and the bitfields' endianess, there is no
    architecture as of Linux-2.6.24-rc4 where the bitfileds' endianess
    does not match the generic endianess. */
@@ -88,8 +89,8 @@ enum after_sb_handler {
 /* KEEP the order, do not delete or insert!
  * Or change the API_VERSION, too. */
 enum ret_codes {
-	RetCodeBase=100,
-	NoError,         // 101 ...
+	RetCodeBase = 100,
+	NoError,         /* 101 ... */
 	LAAlreadyInUse,
 	OAAlreadyInUse,
 	LDNameInvalid,
@@ -117,7 +118,7 @@ enum ret_codes {
 	UnknownMandatoryTag,
 	MinorNotKnown,
 	StateNotAllowed,
-	GotSignal, // EINTR
+	GotSignal, /* EINTR */
 	NoResizeDuringResync,
 	APrimaryNodeNeeded,
 	SyncAfterInvalid,
@@ -129,6 +130,16 @@ enum ret_codes {
 	HaveNoDiskConfig,
 	ProtocolCRequired,
 	VMallocFailed,
+	IntegrityAlgNotAvail,	/* DRBD 8.2 only */
+	IntegrityAlgNotDigest,	/* DRBD 8.2 only */
+	CPUMaskParseFailed,	/* DRBD 8.2 only */
+	CSUMSAlgNotAvail,	/* DRBD 8.2 only */
+	CSUMSAlgNotDigest,	/* DRBD 8.2 only */
+	VERIFYAlgNotAvail,	/* DRBD 8.2 only */
+	VERIFYAlgNotDigest,	/* DRBD 8.2 only */
+	CSUMSResyncRunning,	/* DRBD 8.2 only */
+	VERIFYIsRunning,	/* DRBD 8.2 only */
+	DataOfWrongCurrent,
 
 	/* insert new ones above this line */
 	AfterLastRetCode
@@ -138,12 +149,12 @@ enum ret_codes {
 #define DRBD_PROT_B   2
 #define DRBD_PROT_C   3
 
-typedef enum {
-	Unknown=0,
-	Primary=1,     // role
-	Secondary=2,   // role
-	role_mask=3,
-} drbd_role_t;
+enum drbd_role {
+	Unknown = 0,
+	Primary = 1,     /* role */
+	Secondary = 2,   /* role */
+	role_mask = 3,
+};
 
 /* The order of these constants is important.
  * The lower ones (<WFReportParams) indicate
@@ -154,44 +165,54 @@ typedef enum {
  * Skipped should be < Connected,
  * so writes on a Primary after Skipped sync are not mirrored either ?
  */
-typedef enum {
+enum drbd_conns {
 	StandAlone,
-	Disconnecting,  // Temporal state on the way to StandAlone.
-	Unconnected,    // >= Unconnected -> inc_net() succeeds
-	Timeout,	/// These temporal states are all used on the way
-	BrokenPipe,	/// from >= Connected to Unconnected.
-	NetworkFailure,	/// The 'disconnect reason' states
-	ProtocolError,  ///
-	TearDown,	/// I do not allow to change beween them.
+	Disconnecting,  /* Temporal state on the way to StandAlone. */
+	Unconnected,    /* >= Unconnected -> inc_net() succeeds */
+
+	/* These temporal states are all used on the way
+	 * from >= Connected to Unconnected.
+	 * The 'disconnect reason' states
+	 * I do not allow to change beween them. */
+	Timeout,
+	BrokenPipe,
+	NetworkFailure,
+	ProtocolError,
+	TearDown,
+
 	WFConnection,
-	WFReportParams, // we have a socket
-	Connected,      // we have introduced each other
-	StartingSyncS,  // starting full sync by IOCTL.
-	StartingSyncT,  // stariing full sync by IOCTL.
+	WFReportParams, /* we have a socket */
+	Connected,      /* we have introduced each other */
+	StartingSyncS,  /* starting full sync by IOCTL. */
+	StartingSyncT,  /* stariing full sync by IOCTL. */
 	WFBitMapS,
 	WFBitMapT,
 	WFSyncUUID,
-	SyncSource,     // The distance between original state and pause
-	SyncTarget,     // state must be the same for source and target. (+2)
-	PausedSyncS,    // All SyncStates are tested with this comparison
-	PausedSyncT,    // xx >= SyncSource && xx <= PausedSyncT
-	conn_mask=31
-} drbd_conns_t;
 
-typedef enum {
+	/* All SyncStates are tested with this comparison
+	 * xx >= SyncSource && xx <= PausedSyncT */
+	SyncSource,
+	SyncTarget,
+	PausedSyncS,
+	PausedSyncT,
+	conn_mask = 31
+};
+
+enum drbd_disk_state {
 	Diskless,
 	Attaching,      /* In the process of reading the meta-data */
 	Failed,         /* Becomes Diskless as soon as we told it the peer */
 			/* when >= Failed it is legal to access mdev->bc */
-	Negotiating,    /* Late attaching state, we need to talk to the peer... */
+	Negotiating,    /* Late attaching state, we need to talk to the peer */
 	Inconsistent,
 	Outdated,
 	DUnknown,       /* Only used for the peer, never for myself */
 	Consistent,     /* Might be Outdated, might be UpToDate ... */
 	UpToDate,       /* Only this disk state allows applications' IO ! */
-	disk_mask=15
-} drbd_disks_t;
+	disk_mask = 15
+};
 
+union drbd_state_t {
 /* According to gcc's docs is the ...
  * The order of allocation of bit-fields within a unit (C90 6.5.2.1, C99 6.7.2.1).
  * Determined by ABI.
@@ -200,63 +221,68 @@ typedef enum {
  * the offsets of the bitfields still need to be swapped
  * on different endianess.
  */
-typedef union {
 	struct {
 #if defined(__LITTLE_ENDIAN_BITFIELD)
-		unsigned role : 2 ;   /* 3/4	 primary/secondary/unknown */
-		unsigned peer : 2 ;   /* 3/4	 primary/secondary/unknown */
-		unsigned conn : 5 ;   /* 17/32	 cstates */
-		unsigned disk : 4 ;   /* 8/16	 from Diskless to UpToDate */
-		unsigned pdsk : 4 ;   /* 8/16	 from Diskless to UpToDate */
-		unsigned susp : 1 ;   /* 2/2	 IO suspended  no/yes */
-		unsigned aftr_isp : 1 ; /* isp .. imposed sync pause */
-		unsigned peer_isp : 1 ;
-		unsigned user_isp : 1 ;
-		unsigned _pad : 11;   /* 0	 unused */
+		unsigned role:2 ;   /* 3/4	 primary/secondary/unknown */
+		unsigned peer:2 ;   /* 3/4	 primary/secondary/unknown */
+		unsigned conn:5 ;   /* 17/32	 cstates */
+		unsigned disk:4 ;   /* 8/16	 from Diskless to UpToDate */
+		unsigned pdsk:4 ;   /* 8/16	 from Diskless to UpToDate */
+		unsigned susp:1 ;   /* 2/2	 IO suspended  no/yes */
+		unsigned aftr_isp:1 ; /* isp .. imposed sync pause */
+		unsigned peer_isp:1 ;
+		unsigned user_isp:1 ;
+		unsigned _pad:11;   /* 0	 unused */
 #elif defined(__BIG_ENDIAN_BITFIELD)
-		unsigned _pad : 11;   /* 0	 unused */
-		unsigned user_isp : 1 ;
-		unsigned peer_isp : 1 ;
-		unsigned aftr_isp : 1 ; /* isp .. imposed sync pause */
-		unsigned susp : 1 ;   /* 2/2	 IO suspended  no/yes */
-		unsigned pdsk : 4 ;   /* 8/16	 from Diskless to UpToDate */
-		unsigned disk : 4 ;   /* 8/16	 from Diskless to UpToDate */
-		unsigned conn : 5 ;   /* 17/32	 cstates */
-		unsigned peer : 2 ;   /* 3/4	 primary/secondary/unknown */
-		unsigned role : 2 ;   /* 3/4	 primary/secondary/unknown */
+		unsigned _pad:11;   /* 0	 unused */
+		unsigned user_isp:1 ;
+		unsigned peer_isp:1 ;
+		unsigned aftr_isp:1 ; /* isp .. imposed sync pause */
+		unsigned susp:1 ;   /* 2/2	 IO suspended  no/yes */
+		unsigned pdsk:4 ;   /* 8/16	 from Diskless to UpToDate */
+		unsigned disk:4 ;   /* 8/16	 from Diskless to UpToDate */
+		unsigned conn:5 ;   /* 17/32	 cstates */
+		unsigned peer:2 ;   /* 3/4	 primary/secondary/unknown */
+		unsigned role:2 ;   /* 3/4	 primary/secondary/unknown */
 #else
 # error "this endianess is not supported"
 #endif
 	};
 	unsigned int i;
-} drbd_state_t;
+};
 
-typedef enum {
-	SS_CW_NoNeed=4,
-	SS_CW_Success=3,
-	SS_NothingToDo=2,
-	SS_Success=1,
-	SS_UnknownError=0, // Used to sleep longer in _drbd_request_state
-	SS_TwoPrimaries=-1,
-	SS_NoUpToDateDisk=-2,
-	SS_BothInconsistent=-4,
-	SS_SyncingDiskless=-5,
-	SS_ConnectedOutdates=-6,
-	SS_PrimaryNOP=-7,
-	SS_ResyncRunning=-8,
-	SS_AlreadyStandAlone=-9,
-	SS_CW_FailedByPeer=-10,
-	SS_IsDiskLess=-11,
-	SS_DeviceInUse=-12,
-	SS_NoNetConfig=-13,
-	SS_LowerThanOutdated=-14
-} set_st_err_t;
+enum set_st_err {
+	SS_CW_NoNeed = 4,
+	SS_CW_Success = 3,
+	SS_NothingToDo = 2,
+	SS_Success = 1,
+	SS_UnknownError = 0, /* Used to sleep longer in _drbd_request_state */
+	SS_TwoPrimaries = -1,
+	SS_NoUpToDateDisk = -2,
+	SS_BothInconsistent = -4,
+	SS_SyncingDiskless = -5,
+	SS_ConnectedOutdates = -6,
+	SS_PrimaryNOP = -7,
+	SS_ResyncRunning = -8,
+	SS_AlreadyStandAlone = -9,
+	SS_CW_FailedByPeer = -10,
+	SS_IsDiskLess = -11,
+	SS_DeviceInUse = -12,
+	SS_NoNetConfig = -13,
+	SS_NoVerifyAlg = -14,       /* drbd-8.2 only */
+	SS_NeedConnection = -15,    /* drbd-8.2 only */
+	SS_LowerThanOutdated = -16,
+	SS_NotSupported = -17,      /* drbd-8.2 only */
+	SS_InTransientState = -18,  /* Retry after the next state change */
+	SS_ConcurrentStChg = -19,   /* Concurrent cluster side state change! */
+	SS_AfterLastError = -20,    /* Keep this at bottom */
+};
 
 /* from drbd_strings.c */
-extern const char* conns_to_name(drbd_conns_t);
-extern const char* roles_to_name(drbd_role_t);
-extern const char* disks_to_name(drbd_disks_t);
-extern const char* set_st_err_name(set_st_err_t);
+extern const char *conns_to_name(enum drbd_conns);
+extern const char *roles_to_name(enum drbd_role);
+extern const char *disks_to_name(enum drbd_disk_state);
+extern const char *set_st_err_name(enum set_st_err);
 
 #ifndef BDEVNAME_SIZE
 # define BDEVNAME_SIZE 32
@@ -270,7 +296,7 @@ enum MetaDataFlags {
 	__MDF_ConnectedInd,
 	__MDF_FullSync,
 	__MDF_WasUpToDate,
-	__MDF_PeerOutDated // or less/lower.
+	__MDF_PeerOutDated /* or worse (e.g. invalid). */
 };
 #define MDF_Consistent      (1<<__MDF_Consistent)
 #define MDF_PrimaryInd      (1<<__MDF_PrimaryInd)
@@ -284,9 +310,9 @@ enum UuidIndex {
 	Bitmap,
 	History_start,
 	History_end,
-	UUID_SIZE,      // In the packet we store the number of dirty bits here
-	UUID_FLAGS,     // In the packet we store flags here.
-	EXT_UUID_SIZE   // Everything.
+	UUID_SIZE,      /* nl-packet: number of dirty bits */
+	UUID_FLAGS,     /* nl-packet: flags */
+	EXT_UUID_SIZE   /* Everything. */
 };
 
 #define UUID_JUST_CREATED ((__u64)4)
@@ -299,16 +325,22 @@ enum UuidIndex {
 #define DRBD_MD_INDEX_FLEX_EXT -2
 #define DRBD_MD_INDEX_FLEX_INT -3
 
-// Start of the new netlink/connector stuff
+/* Start of the new netlink/connector stuff */
 
 #define DRBD_NL_CREATE_DEVICE 0x01
 #define DRBD_NL_SET_DEFAULTS  0x02
 
-// The following line should be moved over to linux/connector.h
-// when the time comes
-#define CN_IDX_DRBD			0x4
+/* The following line should be moved over to linux/connector.h
+ * when the time comes */
+#ifndef CN_IDX_DRBD
+# define CN_IDX_DRBD			0x4
+/* Ubuntu's Trapid release defined CN_IDX_DRBD as 0x6 */
+#endif
 #define CN_VAL_DRBD			0x1
 
+/* For searching a vacant cn_idx value */
+#define CN_IDX_STEP			6977
+
 struct drbd_nl_cfg_req {
 	int packet_type;
 	int drbd_minor;
@@ -319,8 +351,8 @@ struct drbd_nl_cfg_req {
 struct drbd_nl_cfg_reply {
 	int packet_type;
 	int minor;
-	int ret_code; // enum ret_code or set_st_err_t
-	unsigned short tag_list[]; // only used with get_* calls
+	int ret_code; /* enum ret_code or set_st_err_t */
+	unsigned short tag_list[]; /* only used with get_* calls */
 };
 
 #endif
