@@ -27,53 +27,65 @@
 
 #include <linux/bitops.h>
 #include <linux/vmalloc.h>
-#include <linux/string.h> // for memset
-#include <linux/seq_file.h> // for seq_printf
+#include <linux/string.h> /* for memset */
+#include <linux/seq_file.h> /* for seq_printf */
 #include "lru_cache.h"
 
 #define STATIC static
 
-// this is developers aid only!
-#define PARANOIA_ENTRY() BUG_ON(test_and_set_bit(__LC_PARANOIA,&lc->flags))
-#define PARANOIA_LEAVE() do { clear_bit(__LC_PARANOIA,&lc->flags); smp_mb__after_clear_bit(); } while (0)
+/* this is developers aid only! */
+#define PARANOIA_ENTRY() BUG_ON(test_and_set_bit(__LC_PARANOIA, &lc->flags))
+#define PARANOIA_LEAVE() do { clear_bit(__LC_PARANOIA, &lc->flags); smp_mb__after_clear_bit(); } while (0)
 #define RETURN(x...)     do { PARANOIA_LEAVE(); return x ; } while (0)
 
+static inline size_t size_of_lc(unsigned int e_count, size_t e_size)
+{
+	return sizeof(struct lru_cache)
+	     + e_count * (e_size + sizeof(struct hlist_head));
+}
+
+static inline void lc_init(struct lru_cache *lc,
+		const size_t bytes, const char *name,
+		const unsigned int e_count, const size_t e_size,
+		void *private_p)
+{
+	struct lc_element *e;
+	unsigned int i;
+
+	memset(lc, 0, bytes);
+	INIT_LIST_HEAD(&lc->in_use);
+	INIT_LIST_HEAD(&lc->lru);
+	INIT_LIST_HEAD(&lc->free);
+	lc->element_size = e_size;
+	lc->nr_elements  = e_count;
+	lc->new_number	 = -1;
+	lc->lc_private   = private_p;
+	lc->name         = name;
+	for (i = 0; i < e_count; i++) {
+		e = lc_entry(lc, i);
+		e->lc_number = LC_FREE;
+		list_add(&e->list, &lc->free);
+		/* memset(,0,) did the rest of init for us */
+	}
+}
+
 /**
  * lc_alloc: allocates memory for @e_count objects of @e_size bytes plus the
  * struct lru_cache, and the hash table slots.
  * returns pointer to a newly initialized lru_cache object with said parameters.
  */
-struct lru_cache* lc_alloc(const char *name, unsigned int e_count,
+struct lru_cache *lc_alloc(const char *name, unsigned int e_count,
 			   size_t e_size, void *private_p)
 {
-	unsigned long bytes;
 	struct lru_cache   *lc;
-	struct lc_element *e;
-	int i;
+	size_t bytes;
 
 	BUG_ON(!e_count);
-	e_size = max(sizeof(struct lc_element),e_size);
-	bytes  = e_size+sizeof(struct hlist_head);
-	bytes *= e_count;
-	bytes += sizeof(struct lru_cache);
-	lc     = vmalloc(bytes);
-	if (lc) {
-		memset(lc, 0, bytes);
-		INIT_LIST_HEAD(&lc->in_use);
-		INIT_LIST_HEAD(&lc->lru);
-		INIT_LIST_HEAD(&lc->free);
-		lc->element_size     = e_size;
-		lc->nr_elements      = e_count;
-		lc->new_number	     = -1;
-		lc->lc_private       = private_p;
-		lc->name             = name;
-		for(i=0;i<e_count;i++) {
-			e = lc_entry(lc,i);
-			e->lc_number = LC_FREE;
-			list_add(&e->list,&lc->free);
-			// memset(,0,) did the rest of init for us
-		}
-	}
+	e_size = max(sizeof(struct lc_element), e_size);
+	bytes = size_of_lc(e_count, e_size);
+	lc = vmalloc(bytes);
+	if (lc)
+		lc_init(lc, bytes, name, e_count, e_size, private_p);
 	return lc;
 }
 
@@ -81,26 +93,38 @@ struct lru_cache* lc_alloc(const char *name, unsigned int e_count,
  * lc_free: Frees memory allocated by lc_alloc.
  * @lc: The lru_cache object
  */
-void lc_free(struct lru_cache* lc)
+void lc_free(struct lru_cache *lc)
 {
 	vfree(lc);
 }
 
-size_t	lc_printf_stats(struct seq_file *seq, struct lru_cache* lc)
+/**
+ * lc_reset: does a full reset for @lc and the hash table slots.
+ * It is roughly the equivalent of re-allocating a fresh lru_cache object,
+ * basically a short cut to lc_free(lc); lc = lc_alloc(...);
+ */
+void lc_reset(struct lru_cache *lc)
+{
+	lc_init(lc, size_of_lc(lc->nr_elements, lc->element_size), lc->name,
+			lc->nr_elements, lc->element_size, lc->lc_private);
+}
+
+size_t	lc_printf_stats(struct seq_file *seq, struct lru_cache *lc)
 {
 	/* NOTE:
 	 * total calls to lc_get are
-	 * starving + hits + misses
-	 * misses include "dirty" count (update from an other thread in progress)
-	 * and "changed", when this in fact lead to an successful update of the cache.
+	 * (starving + hits + misses)
+	 * misses include "dirty" count (update from an other thread in
+	 * progress) and "changed", when this in fact lead to an successful
+	 * update of the cache.
 	 */
-	return seq_printf(seq,"\t%s: used:%u/%u "
+	return seq_printf(seq, "\t%s: used:%u/%u "
 		"hits:%lu misses:%lu starving:%lu dirty:%lu changed:%lu\n",
 		lc->name, lc->used, lc->nr_elements,
 		lc->hits, lc->misses, lc->starving, lc->dirty, lc->changed);
 }
 
-static unsigned int lc_hash_fn(struct lru_cache* lc, unsigned int enr)
+static unsigned int lc_hash_fn(struct lru_cache *lc, unsigned int enr)
 {
 	return enr % lc->nr_elements;
 }
@@ -112,7 +136,7 @@ static unsigned int lc_hash_fn(struct lru_cache* lc, unsigned int enr)
  * @lc: The lru_cache object
  * @enr: element number
  */
-struct lc_element* lc_find(struct lru_cache* lc, unsigned int enr)
+struct lc_element *lc_find(struct lru_cache *lc, unsigned int enr)
 {
 	struct hlist_node *n;
 	struct lc_element *e;
@@ -120,20 +144,22 @@ struct lc_element* lc_find(struct lru_cache* lc, unsigned int enr)
 	BUG_ON(!lc);
 	BUG_ON(!lc->nr_elements);
 	hlist_for_each_entry(e, n, lc->slot + lc_hash_fn(lc, enr), colision) {
-		if (e->lc_number == enr) return e;
+		if (e->lc_number == enr)
+			return e;
 	}
 	return NULL;
 }
 
-STATIC struct lc_element * lc_evict(struct lru_cache* lc)
+STATIC struct lc_element *lc_evict(struct lru_cache *lc)
 {
 	struct list_head  *n;
 	struct lc_element *e;
 
-	if (list_empty(&lc->lru)) return 0;
+	if (list_empty(&lc->lru))
+		return NULL;
 
-	n=lc->lru.prev;
-	e=list_entry(n, struct lc_element,list);
+	n = lc->lru.prev;
+	e = list_entry(n, struct lc_element, list);
 
 	list_del(&e->list);
 	hlist_del(&e->colision);
@@ -147,34 +173,37 @@ STATIC struct lc_element * lc_evict(struct lru_cache* lc)
  * @lc: The lru_cache object
  * @e: The element to remove
  */
-void lc_del(struct lru_cache* lc, struct lc_element *e)
+void lc_del(struct lru_cache *lc, struct lc_element *e)
 {
-	// FIXME what to do with refcnt != 0 ?
+	/* FIXME what to do with refcnt != 0 ? */
 	PARANOIA_ENTRY();
 	BUG_ON(e->refcnt);
 	list_del(&e->list);
 	hlist_del_init(&e->colision);
 	e->lc_number = LC_FREE;
 	e->refcnt = 0;
-	list_add(&e->list,&lc->free);
+	list_add(&e->list, &lc->free);
 	RETURN();
 }
 
-STATIC struct lc_element* lc_get_unused_element(struct lru_cache* lc)
+STATIC struct lc_element *lc_get_unused_element(struct lru_cache *lc)
 {
 	struct list_head *n;
 
-	if (list_empty(&lc->free)) return lc_evict(lc);
+	if (list_empty(&lc->free))
+		return lc_evict(lc);
 
-	n=lc->free.next;
+	n = lc->free.next;
 	list_del(n);
-	return list_entry(n, struct lc_element,list);
+	return list_entry(n, struct lc_element, list);
 }
 
-STATIC int lc_unused_element_available(struct lru_cache* lc)
+STATIC int lc_unused_element_available(struct lru_cache *lc)
 {
-	if (!list_empty(&lc->free)) return 1; // something on the free list
-	if (!list_empty(&lc->lru)) return 1;  // something to evict
+	if (!list_empty(&lc->free))
+		return 1; /* something on the free list */
+	if (!list_empty(&lc->lru))
+		return 1;  /* something to evict */
 
 	return 0;
 }
@@ -207,7 +236,7 @@ STATIC int lc_unused_element_available(struct lru_cache* lc)
  * @lc: The lru_cache object
  * @enr: element number
  */
-struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
+struct lc_element *lc_get(struct lru_cache *lc, unsigned int enr)
 {
 	struct lc_element *e;
 
@@ -215,7 +244,7 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	BUG_ON(!lc->nr_elements);
 
 	PARANOIA_ENTRY();
-	if ( lc->flags & LC_STARVING ) {
+	if (lc->flags & LC_STARVING) {
 		++lc->starving;
 		RETURN(NULL);
 	}
@@ -223,8 +252,9 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	e = lc_find(lc, enr);
 	if (e) {
 		++lc->hits;
-		if( e->refcnt++ == 0) lc->used++;
-		list_move(&e->list,&lc->in_use); // Not evictable...
+		if (e->refcnt++ == 0)
+			lc->used++;
+		list_move(&e->list, &lc->in_use); /* Not evictable... */
 		RETURN(e);
 	}
 
@@ -233,8 +263,8 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	/* In case there is nothing available and we can not kick out
 	 * the LRU element, we have to wait ...
 	 */
-	if(!lc_unused_element_available(lc)) {
-		__set_bit(__LC_STARVING,&lc->flags);
+	if (!lc_unused_element_available(lc)) {
+		__set_bit(__LC_STARVING, &lc->flags);
 		RETURN(NULL);
 	}
 
@@ -242,7 +272,7 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	 * which then is replaced.
 	 * we need to update the cache; serialize on lc->flags & LC_DIRTY
 	 */
-	if (test_and_set_bit(__LC_DIRTY,&lc->flags)) {
+	if (test_and_set_bit(__LC_DIRTY, &lc->flags)) {
 		++lc->dirty;
 		RETURN(NULL);
 	}
@@ -250,7 +280,7 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	e = lc_get_unused_element(lc);
 	BUG_ON(!e);
 
-	clear_bit(__LC_STARVING,&lc->flags);
+	clear_bit(__LC_STARVING, &lc->flags);
 	BUG_ON(++e->refcnt != 1);
 	lc->used++;
 
@@ -264,7 +294,7 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
  * but only gets a new reference on an existing element.
  * you either get the requested element, or NULL.
  */
-struct lc_element* lc_try_get(struct lru_cache* lc, unsigned int enr)
+struct lc_element *lc_try_get(struct lru_cache *lc, unsigned int enr)
 {
 	struct lc_element *e;
 
@@ -272,7 +302,7 @@ struct lc_element* lc_try_get(struct lru_cache* lc, unsigned int enr)
 	BUG_ON(!lc->nr_elements);
 
 	PARANOIA_ENTRY();
-	if ( lc->flags & LC_STARVING ) {
+	if (lc->flags & LC_STARVING) {
 		++lc->starving;
 		RETURN(NULL);
 	}
@@ -280,29 +310,31 @@ struct lc_element* lc_try_get(struct lru_cache* lc, unsigned int enr)
 	e = lc_find(lc, enr);
 	if (e) {
 		++lc->hits;
-		if( e->refcnt++ == 0) lc->used++;
-		list_move(&e->list,&lc->in_use); // Not evictable...
+		if (e->refcnt++ == 0)
+			lc->used++;
+		list_move(&e->list, &lc->in_use); /* Not evictable... */
 	}
 	RETURN(e);
 }
 
-void lc_changed(struct lru_cache* lc, struct lc_element* e)
+void lc_changed(struct lru_cache *lc, struct lc_element *e)
 {
 	PARANOIA_ENTRY();
 	BUG_ON(e != lc->changing_element);
 	++lc->changed;
 	e->lc_number = lc->new_number;
-	list_add(&e->list,&lc->in_use);
-	hlist_add_head( &e->colision, lc->slot + lc_hash_fn(lc, lc->new_number) );
+	list_add(&e->list, &lc->in_use);
+	hlist_add_head(&e->colision,
+		lc->slot + lc_hash_fn(lc, lc->new_number));
 	lc->changing_element = NULL;
 	lc->new_number = -1;
-	clear_bit(__LC_DIRTY,&lc->flags);
+	clear_bit(__LC_DIRTY, &lc->flags);
 	smp_mb__after_clear_bit();
 	PARANOIA_LEAVE();
 }
 
 
-unsigned int lc_put(struct lru_cache* lc, struct lc_element* e)
+unsigned int lc_put(struct lru_cache *lc, struct lc_element *e)
 {
 	BUG_ON(!lc);
 	BUG_ON(!lc->nr_elements);
@@ -311,10 +343,11 @@ unsigned int lc_put(struct lru_cache* lc, struct lc_element* e)
 	PARANOIA_ENTRY();
 	BUG_ON(e->refcnt == 0);
 	BUG_ON(e == lc->changing_element);
-	if ( --e->refcnt == 0) {
-		list_move(&e->list,&lc->lru); // move it to the front of LRU.
+	if (--e->refcnt == 0) {
+		/* move it to the front of LRU. */
+		list_move(&e->list, &lc->lru);
 		lc->used--;
-		clear_bit(__LC_STARVING,&lc->flags);
+		clear_bit(__LC_STARVING, &lc->flags);
 		smp_mb__after_clear_bit();
 	}
 	RETURN(e->refcnt);
@@ -328,41 +361,42 @@ unsigned int lc_put(struct lru_cache* lc, struct lc_element* e)
  * @enr: element number
  * @index: The elements' position in the cache
  */
-void lc_set(struct lru_cache* lc, unsigned int enr, int index)
+void lc_set(struct lru_cache *lc, unsigned int enr, int index)
 {
 	struct lc_element *e;
 
-	if ( index < 0 || index >= lc->nr_elements ) return;
+	if (index < 0 || index >= lc->nr_elements)
+		return;
 
-	e = lc_entry(lc,index);
+	e = lc_entry(lc, index);
 	e->lc_number = enr;
 
 	hlist_del_init(&e->colision);
-	hlist_add_head( &e->colision, lc->slot + lc_hash_fn(lc,enr) );
+	hlist_add_head(&e->colision, lc->slot + lc_hash_fn(lc, enr));
 	list_move(&e->list, e->refcnt ? &lc->in_use : &lc->lru);
 }
 
-#if 0
+#ifdef DRBD_DUMP_RESYNC_DETAIL
 /**
  * lc_dump: Dump a complete LRU cache to seq in textual form.
  */
-void lc_dump(struct lru_cache* lc, struct seq_file *seq, char* utext,
-	     void (*detail) (struct seq_file *, struct lc_element *) )
+void lc_dump(struct lru_cache *lc, struct seq_file *seq, char *utext,
+	     void (*detail) (struct seq_file *, struct lc_element *))
 {
 	unsigned int nr_elements = lc->nr_elements;
 	struct lc_element *e;
 	int i;
 
-	seq_printf(seq,"\tnn: lc_number refcnt %s\n ",utext);
-	for(i=0;i<nr_elements;i++) {
-		e = lc_entry(lc,i);
-		if( e->lc_number == LC_FREE ) {
-			seq_printf(seq,"\t%2d: FREE\n",i );
+	seq_printf(seq, "\tnn: lc_number refcnt %s\n ", utext);
+	for (i = 0; i < nr_elements; i++) {
+		e = lc_entry(lc, i);
+		if (e->lc_number == LC_FREE) {
+			seq_printf(seq, "\t%2d: FREE\n", i);
 		} else {
-			seq_printf(seq,"\t%2d: %4u %4u    ", i,
+			seq_printf(seq, "\t%2d: %4u %4u    ", i,
 				   e->lc_number,
-				   e->refcnt );
-			detail(seq,e);
+				   e->refcnt);
+			detail(seq, e);
 		}
 	}
 }
