@@ -68,8 +68,17 @@ char* progname;
 struct adm_cmd {
   const char* name;
   int (* function)(struct d_resource*,const char* );
+  /* which level this command is for.
+   * 0: don't show this command, ever
+   * 1: normal administrative commands, shown in normal help
+   * 2-4: shown on "drbdadm hidden-commands"
+   * 2: usefull for shell scripts
+   * 3: callbacks potentially called from kernel module on certain events
+   * 4: advanced, experts and developers only */
   unsigned int show_in_usage     :3;
+  /* if set, command requires an explicit resource name */
   unsigned int res_name_required :1;
+  /* error out if the ip specified is not available/active now */
   unsigned int verify_ips        :1;
 };
 
@@ -99,6 +108,7 @@ static int adm_wait_c(struct d_resource* ,const char* );
 static int adm_wait_ci(struct d_resource* ,const char* );
 static int sh_nop(struct d_resource* ,const char* );
 static int sh_resources(struct d_resource* ,const char* );
+static int sh_resource(struct d_resource* ,const char* );
 static int sh_mod_parms(struct d_resource* ,const char* );
 static int sh_dev(struct d_resource* ,const char* );
 static int sh_ip(struct d_resource* ,const char* );
@@ -110,6 +120,7 @@ static int admm_generic(struct d_resource* ,const char* );
 static int adm_khelper(struct d_resource* ,const char* );
 static int adm_generic_b(struct d_resource* ,const char* );
 static int hidden_cmds(struct d_resource* ,const char* );
+static int adm_outdate(struct d_resource* ,const char* );
 
 static char* get_opt_val(struct d_option*,const char*,char*);
 
@@ -197,54 +208,108 @@ struct option admopt[] = {
   { 0,              0,                0, 0   }
 };
 
+/* DRBD adm_cmd flags shortcuts,
+ * to avoid merge conflicts and unreadable diffs
+ * when we add the next flag */
+
+#define DRBD_acf1_default		\
+	.show_in_usage = 1,		\
+	.res_name_required = 1,		\
+	.verify_ips = 0,
+
+#define DRBD_acf1_defnet		\
+	.show_in_usage = 1,		\
+	.res_name_required = 1,		\
+	.verify_ips = 1,
+
+#define DRBD_acf3_handler		\
+	.show_in_usage = 3,		\
+	.res_name_required = 1,		\
+	.verify_ips = 0,
+
+#define DRBD_acf4_advanced		\
+	.show_in_usage = 4,		\
+	.res_name_required = 1,		\
+	.verify_ips = 0,
+
+#define DRBD_acf1_dump			\
+	.show_in_usage = 1,		\
+	.res_name_required = 1,		\
+	.verify_ips = 1,
+
+#define DRBD_acf2_shell			\
+	.show_in_usage = 2,		\
+	.res_name_required = 1,		\
+	.verify_ips = 0,
+
+#define DRBD_acf2_gen_shell		\
+	.show_in_usage = 2,		\
+	.res_name_required = 0,		\
+	.verify_ips = 0,
+
 struct adm_cmd cmds[] = {
-/*   name, function,                  show, needs res, verify_ips */
-  { "attach",            adm_attach,    1,1,0 },
-  { "detach",            adm_generic_s, 1,1,0 },
-  { "connect",           adm_connect,   1,1,1 },
-  { "disconnect",        adm_generic_s, 1,1,0 },
-  { "up",                adm_up,        1,1,1 },
-  { "down",              adm_generic_s, 1,1,0 },
-  { "primary",           adm_generic_l, 1,1,0 },
-  { "secondary",         adm_generic_l, 1,1,0 },
-  { "invalidate",        adm_generic_b, 1,1,0 },
-  { "invalidate-remote", adm_generic_l, 1,1,1 },
-  { "outdate",           adm_generic_b, 1,1,0 },
-  { "resize",            adm_resize,    1,1,1 },
-  { "syncer",            adm_syncer,    1,1,1 },
-  { "pause-sync",        adm_generic_s, 1,1,1 },
-  { "resume-sync",       adm_generic_s, 1,1,1 },
-  { "adjust",            adm_adjust,    1,1,1 },
-  { "wait-connect",      adm_wait_c,    1,1,1 },
-  { "state",             adm_generic_s, 1,1,0 },
-  { "cstate",            adm_generic_s, 1,1,0 },
-  { "dstate",            adm_generic_b, 1,1,0 },
-  { "dump",              adm_dump,      1,1,1 },
-  { "dump-xml",          adm_dump_xml,  1,1,1 },
-  { "create-md",         adm_create_md, 1,1,0 },
-  { "show-gi",           adm_generic_b, 1,1,0 },
-  { "get-gi",            adm_generic_b, 1,1,0 },
-  { "dump-md",           admm_generic,  1,1,0 },
-  { "wait-con-int",      adm_wait_ci,   1,0,1 },
-  { "hidden-commands",   hidden_cmds,   1,0,0 },
-  { "sh-nop",            sh_nop,        2,0,0 },
-  { "sh-resources",      sh_resources,  2,0,0 },
-  { "sh-mod-parms",      sh_mod_parms,  2,0,0 },
-  { "sh-dev",            sh_dev,        2,1,0 },
-  { "sh-ll-dev",         sh_ll_dev,     2,1,0 },
-  { "sh-md-dev",         sh_md_dev,     2,1,0 },
-  { "sh-md-idx",         sh_md_idx,     2,1,0 },
-  { "sh-ip",             sh_ip,         0,1,0 },
-  { "sh-b-pri",          sh_b_pri,      2,1,0 },
-  { "pri-on-incon-degr", adm_khelper,   3,1,0 },
-  { "pri-lost-after-sb", adm_khelper,   3,1,0 },
-  { "outdate-peer",      adm_khelper,   3,1,0 },
-  { "local-io-error",    adm_khelper,   3,1,0 },
-  { "pri-lost",          adm_khelper,   3,1,0 },
-  { "split-brain",       adm_khelper,   3,1,0 },
-  { "set-gi",            admm_generic,  4,1,0 },
-  { "suspend-io",        adm_generic_s, 4,1,0 },
-  { "resume-io",         adm_generic_s, 4,1,0 },
+        /*  name, function, flags
+	 *  sort order:
+	 *  - normal config commands,
+	 *  - normal meta data manipulation
+	 *  - sh-*
+	 *  - handler
+	 *  - advanced
+	 ***/
+        { "attach",                adm_attach,      DRBD_acf1_default   },
+        { "detach",                adm_generic_l,   DRBD_acf1_default   },
+        { "connect",               adm_connect,     DRBD_acf1_defnet    },
+        { "disconnect",            adm_generic_s,   DRBD_acf1_default   },
+        { "up",                    adm_up,          DRBD_acf1_defnet    },
+        { "down",                  adm_generic_l,   DRBD_acf1_default   },
+        { "primary",               adm_generic_l,   DRBD_acf1_default   },
+        { "secondary",             adm_generic_l,   DRBD_acf1_default   },
+        { "invalidate",            adm_generic_b,   DRBD_acf1_default   },
+        { "invalidate-remote",     adm_generic_l,   DRBD_acf1_defnet    },
+        { "outdate",               adm_outdate,     DRBD_acf1_default   },
+        { "resize",                adm_resize,      DRBD_acf1_defnet    },
+        { "syncer",                adm_syncer,      DRBD_acf1_defnet    },
+        { "pause-sync",            adm_generic_s,   DRBD_acf1_defnet    },
+        { "resume-sync",           adm_generic_s,   DRBD_acf1_defnet    },
+        { "adjust",                adm_adjust,      DRBD_acf1_defnet    },
+        { "wait-connect",          adm_wait_c,      DRBD_acf1_defnet    },
+        { "wait-con-int",          adm_wait_ci,
+		.show_in_usage = 1, .verify_ips = 1, },
+        { "state",                 adm_generic_s,   DRBD_acf1_default   },
+        { "cstate",                adm_generic_s,   DRBD_acf1_default   },
+        { "dstate",                adm_generic_b,   DRBD_acf1_default   },
+
+        { "dump",                  adm_dump,        DRBD_acf1_dump      },
+        { "dump-xml",              adm_dump_xml,    DRBD_acf1_dump      },
+
+        { "create-md",             adm_create_md,   DRBD_acf1_default   },
+        { "show-gi",               adm_generic_b,   DRBD_acf1_default   },
+        { "get-gi",                adm_generic_b,   DRBD_acf1_default   },
+        { "dump-md",               admm_generic,    DRBD_acf1_default   },
+        { "wipe-md",               admm_generic,    DRBD_acf1_default   },
+        { "hidden-commands",       hidden_cmds,     .show_in_usage = 1, },
+
+        { "sh-nop",                sh_nop,          DRBD_acf2_gen_shell },
+        { "sh-resources",          sh_resources,    DRBD_acf2_gen_shell },
+        { "sh-resource",           sh_resource,     DRBD_acf2_shell     },
+        { "sh-mod-parms",          sh_mod_parms,    DRBD_acf2_gen_shell },
+        { "sh-dev",                sh_dev,          DRBD_acf2_shell     },
+        { "sh-ll-dev",             sh_ll_dev,       DRBD_acf2_shell     },
+        { "sh-md-dev",             sh_md_dev,       DRBD_acf2_shell     },
+        { "sh-md-idx",             sh_md_idx,       DRBD_acf2_shell     },
+        { "sh-ip",                 sh_ip,           DRBD_acf2_shell     },
+        { "sh-b-pri",              sh_b_pri,        DRBD_acf2_shell     },
+
+        { "pri-on-incon-degr",     adm_khelper,     DRBD_acf3_handler   },
+        { "pri-lost-after-sb",     adm_khelper,     DRBD_acf3_handler   },
+        { "outdate-peer",          adm_khelper,     DRBD_acf3_handler   },
+        { "local-io-error",        adm_khelper,     DRBD_acf3_handler   },
+        { "pri-lost",              adm_khelper,     DRBD_acf3_handler   },
+        { "split-brain",           adm_khelper,     DRBD_acf3_handler   },
+
+        { "suspend-io",            adm_generic_s,   DRBD_acf4_advanced  },
+        { "resume-io",             adm_generic_s,   DRBD_acf4_advanced  },
+        { "set-gi",                admm_generic,    DRBD_acf4_advanced  },
 };
 
 /*** These functions are used to the print the config ***/
@@ -505,6 +570,13 @@ static int sh_resources(struct d_resource* ignored __attribute((unused)),const c
   return 0;
 }
 
+static int sh_resource(struct d_resource* res,const char* unused __attribute((unused)))
+{
+  printf("%s\n",res->name);
+
+  return 0;
+}
+
 static int sh_dev(struct d_resource* res,const char* unused __attribute((unused)))
 {
   printf("%s\n",res->me->device);
@@ -548,11 +620,19 @@ static int sh_md_idx(struct d_resource* res,const char* unused __attribute((unus
 static int sh_b_pri(struct d_resource *res,const char* unused __attribute((unused)))
 {
   char* val;
+  int i, rv;
 
   val = get_opt_val(res->startup_options, "become-primary-on", NULL);
   if ( val && ( !strcmp(val,nodeinfo.nodename) ||
 		!strcmp(val,"both") ) ) {
-    return adm_generic_s(res,"primary");
+    /* Opon connect resync starts, and both sides become primary at the same time.
+       One's try might be declined since an other state transition happens. Retry. */
+    for (i=0; i<5; i++) {
+      rv = adm_generic_s(res,"primary");
+      if (rv == 0) return rv;
+      sleep(1);
+    }
+    return rv;
   }
   return 0;
 }
@@ -682,13 +762,11 @@ static inline const char* shell_escape(char* s)
 	if (s == NULL)
 		return s;
 
-	while (1) {
+	while (*s) {
 		if (buffer + sizeof(buffer) < c+2)
 			break;
 
 		switch(*s) {
-		case 0: /* terminator */
-			break;
 		/* set of 'clean' characters */
 		case '%': case '+': case '-': case '.': case '/':
 		case '0' ... '9':
@@ -711,6 +789,7 @@ pid_t m_system(char** argv, int flags, struct d_resource *res)
 {
   pid_t pid;
   int status,rv=-1;
+  int timeout = 0;
   char **cmdline = argv;
 
   struct sigaction so;
@@ -743,7 +822,6 @@ pid_t m_system(char** argv, int flags, struct d_resource *res)
   }
 
   if( flags & SLEEPS_FINITE ) {
-    int timeout;
     sigaction(SIGALRM,&sa,&so);
     alarm_raised=0;
     switch(flags & SLEEPS_MASK) {
@@ -765,8 +843,10 @@ pid_t m_system(char** argv, int flags, struct d_resource *res)
     if (waitpid(pid, &status, 0) == -1) {
       if (errno != EINTR) break;
       if (alarm_raised) {
-	fprintf(stderr,"Child process does not terminate!\nExiting.\n");
-	exit(E_exec_error);
+	alarm(0);
+	sigaction(SIGALRM,&so,NULL);
+	rv = 0x100;
+	break;
       } else {
 	fprintf(stderr,"logic bug in %s:%d\n",__FILE__,__LINE__);
 	exit(E_exec_error);
@@ -779,16 +859,19 @@ pid_t m_system(char** argv, int flags, struct d_resource *res)
     }
   }
 
-  if( flags & SLEEPS_FINITE ) {
-    alarm(0);
-    sigaction(SIGALRM,&so,NULL);
-    if(rv >= 10 && !(flags & (DONT_REPORT_FAILED|SUPRESS_STDERR))) {
+  if (flags & SLEEPS_FINITE) {
+    if (rv >= 10 && !(flags & (DONT_REPORT_FAILED|SUPRESS_STDERR))) {
       fprintf(stderr,"Command '");
-      while(*argv) {
-	fprintf(stderr,"%s",*argv++);
-	if (*argv) fputc(' ',stderr);
+      for (cmdline = argv; *cmdline; cmdline++) {
+	fprintf(stderr, "%s", *cmdline);
+	if (cmdline[1]) fputc(' ',stderr);
+      }
+      if (alarm_raised) {
+	fprintf(stderr,"' did not terminate within %u seconds\n", timeout);
+	exit(E_exec_error);
+      } else {
+	fprintf(stderr,"' terminated with exit code %d\n",rv);
       }
-      fprintf(stderr,"' terminated with exit code %d\n",rv);
     }
   }
 
@@ -943,15 +1026,37 @@ int adm_generic_l(struct d_resource* res,const char* cmd)
   return adm_generic(res,cmd,SLEEPS_LONG);
 }
 
+static int adm_outdate(struct d_resource* res,const char* cmd)
+{
+  int rv;
+
+  rv=adm_generic(res,cmd,SLEEPS_SHORT|SUPRESS_STDERR);
+  /* special cases for outdate:
+   * 17: drbdsetup outdate, but is primary and thus cannot be outdated.
+   *  5: drbdsetup outdate, and is inconsistent or worse anyways. */
+  if (rv == 17)
+    return rv;
+
+  if (rv == 5) {
+    /* That might mean it is diskless. */
+    rv = admm_generic(res,cmd);
+    if (rv) rv = 5;
+    return rv;
+  }
+
+  if (rv || dry_run) {
+    rv = admm_generic(res,cmd);
+  }
+  return rv;
+}
+
 static int adm_generic_b(struct d_resource* res,const char* cmd)
 {
   int rv;
 
   rv=adm_generic(res,cmd,SLEEPS_SHORT|SUPRESS_STDERR);
-  if(rv == 17) return rv;
-  // 17 returned by drbdsetup outdate, if it is already primary.
 
-  if( rv || dry_run ) {
+  if (rv || dry_run) {
     rv = admm_generic(res,cmd);
   }
   return rv;
@@ -1561,22 +1666,27 @@ int check_uniq(const char* what, const char *fmt, ...)
 int sanity_check_abs_cmd(char* cmd_name)
 {
   struct stat sb;
-  int err;
 
-  if ( (err=stat(cmd_name,&sb)) ) {
+  if (stat(cmd_name,&sb)) {
+    /* If stat fails, just ignore this sanity check,
+     * we are still iterating over $PATH probably. */
     return 0;
   }
 
   if(!sb.st_mode&S_ISUID || sb.st_mode&S_IXOTH || sb.st_gid==0) {
-    fprintf(stderr,"WARN:\n"
-	    "  You are using the 'drbd-peer-outdater' as outdate-peer program.\n"
-	    "  If you use that mechanism the dopd heartbeat plugin program needs\n"
-	    "  to be able to call drbdsetup and drbdmeta with root privileges.\n\n"
-	    "  You need to fix this with these commands:\n"
-	    "  chgrp haclient %s\n"
-	    "  chmod o-x %s\n"
-	    "  chmod u+s %s\n\n\n",
-	    cmd_name,cmd_name,cmd_name);
+    static int did_header = 0;
+    if (!did_header)
+      fprintf(stderr,
+	"WARN:\n"
+	"  You are using the 'drbd-peer-outdater' as outdate-peer program.\n"
+	"  If you use that mechanism the dopd heartbeat plugin program needs\n"
+	"  to be able to call drbdsetup and drbdmeta with root privileges.\n\n"
+	"  You need to fix this with these commands:\n");
+    did_header = 1;
+    fprintf(stderr,
+	"  chgrp haclient %s\n"
+	"  chmod o-x %s\n"
+	"  chmod u+s %s\n\n", cmd_name, cmd_name, cmd_name);
   }
   return 1;
 }
@@ -1605,14 +1715,53 @@ void sanity_check_cmd(char* cmd_name)
   }
 }
 
+/* if the config file is not readable by haclient,
+ * dopd cannot work.
+ * NOTE: we assume that any gid != 0 will be the group dopd will run as,
+ * typically haclient. */
+void sanity_check_conf(char *c)
+{
+	struct stat sb;
+
+	/* if we cannot stat the config file,
+	* we have other things to worry about. */
+	if (stat(c,&sb))
+		return;
+
+	/* permissions are funny: if it is world readable,
+	 * but not group readable, and it belongs to my group,
+	 * I am denied access.
+	 * For the file to be readable by dopd (hacluster:haclient),
+	 * it is not enough to be world readable. */
+
+	/* ok if world readable, and NOT group haclient (see NOTE above) */
+	if (sb.st_mode & S_IROTH && sb.st_gid == 0)
+		return;
+
+	/* ok if group readable, and group haclient (see NOTE above) */
+	if (sb.st_mode & S_IRGRP && sb.st_gid != 0)
+		return;
+
+	fprintf(stderr,
+		"WARN:\n"
+		"  You are using the 'drbd-peer-outdater' as outdate-peer program.\n"
+		"  If you use that mechanism the dopd heartbeat plugin program needs\n"
+		"  to be able to read the drbd.config file.\n\n"
+		"  You need to fix this with these commands:\n"
+		"  chgrp haclient %s\n"
+		"  chmod g+r %s\n\n", c, c);
+}
+
 void sanity_check_perm()
 {
   static int checked=0;
-
-  if(checked) return;
+  if (checked)
+	  return;
 
   sanity_check_cmd(drbdsetup);
   sanity_check_cmd(drbdmeta);
+  sanity_check_conf(config_file);
+  checked = 1;
 }
 
 void validate_resource(struct d_resource * res)
@@ -1668,9 +1817,8 @@ void validate_resource(struct d_resource * res)
       }
     }
   }
-  if (res->me && res->peer) {
-    verify_ips(res);
-  }
+  /* IP verification (check for existence)
+   * moved to just before command execution */
 
   if ( (opt = find_opt(res->handlers, "outdate-peer")) ) {
     if(strstr(opt->value,"drbd-peer-outdater")) sanity_check_perm();
@@ -1957,7 +2105,16 @@ int main(int argc, char** argv)
 	}
       }
 
+      /* either no resorce arguments at all,
+       * but command is dump / dump-xml, so implitict "all",
+       * or an explicit "all" argument is given */
       if ( optind==argc || !strcmp(argv[optind],"all") ) {
+	/* verify ips first, for all of them */
+        for_each_resource(res,tmp,config) {
+	  verify_ips(res);
+	}
+	if (!is_dump && !config_valid)
+	  exit(E_config_invalid);
         if (is_dump) {
           if (is_dump_xml) {
             printf("<config file=\"%s\">\n", config_file); ++indent;
@@ -1986,6 +2143,9 @@ int main(int argc, char** argv)
 	    fprintf(stderr,"'%s' not defined in your config.\n",argv[i]);
 	    exit(E_usage);
 	  }
+	  verify_ips(res);
+	  if (!is_dump && !config_valid)
+            exit(E_config_invalid);
 	  if( (rv=cmd->function(res,cmd->name)) >= 20 ) {
 	    fprintf(stderr,"drbdadm aborting\n");
 	    exit(rv);
