@@ -49,6 +49,8 @@ enum range_checks
 enum yytokentype {
 	TK_GLOBAL = 258,
 	TK_RESOURCE,
+
+	/* 260 */
 	TK_ON,
 	TK_NET,
 	TK_DISK,
@@ -58,19 +60,21 @@ enum yytokentype {
 	TK_DISABLE_IP_VERIFICATION,
 	TK_DIALOG_REFRESH,
 	TK_PROTOCOL,
-	TK_HANDLER, // 270
+	TK_HANDLER,
+
+	/* 270 */
 	TK_COMMON,
 	TK_ADDRESS,
 	TK_DEVICE,
 	TK_META_DISK,
 	TK_FLEX_META_DISK,
 	TK_MINOR_COUNT,
-
 	TK_IPADDR,
 	TK_INTEGER,
 	TK_STRING,
-	TK_ELSE, // 280
+	TK_ELSE,
 
+	/* 280 */
 	TK_DISK_SWITCH,
 	TK_DISK_OPTION,
 	TK_NET_SWITCH,
@@ -80,7 +84,9 @@ enum yytokentype {
 	TK_STARTUP_SWITCH,
 	TK_STARTUP_OPTION,
 	TK_HANDLER_OPTION,
-	TK_USAGE_COUNT,  // 290
+	TK_USAGE_COUNT,
+
+	/* 290 */
 	TK_ASK,
 	TK_YES,
 	TK_NO,
@@ -88,7 +94,18 @@ enum yytokentype {
 	TK__MAJOR,
 	TK__MINOR,
 	TK__THIS_HOST,
-	TK__REMOTE_HOST
+	TK__REMOTE_HOST,
+	TK_PROXY,
+	TK_INSIDE,
+
+	/* 300 */
+	TK_OUTSIDE,
+	TK_MEMLIMIT,
+	TK_PROXY_OPTION,
+	TK_PROXY_SWITCH,
+	TK_ERR_STRING_TOO_LONG,
+	TK_ERR_DQSTRING_TOO_LONG,
+	TK_ERR_DQSTRING,
 };
 
 typedef struct YYSTYPE {
