@@ -45,6 +45,7 @@
 #include <sys/time.h>
 #include <time.h>
 
+#define __bitwise /* Build-workaround for broken RHEL4 kernels (2.6.9_78.0.1) */
 #include <linux/netlink.h>
 #include <linux/connector.h>
 
@@ -60,9 +61,9 @@
 #endif
 
 enum usage_type {
-  BRIEF,
-  FULL,
-  XML,
+	BRIEF,
+	FULL,
+	XML,
 };
 
 struct drbd_tag_list {
@@ -124,7 +125,7 @@ struct drbd_cmd {
 		} gp; // for generic_get_cmd, get_usage
 		struct {
 			struct option *options;
-			int (*proc_event)(unsigned int, int, 
+			int (*proc_event)(unsigned int, int,
 					  struct drbd_nl_cfg_reply *);
 		} ep; // for events_cmd, events_usage
 	};
@@ -181,6 +182,7 @@ int conv_protocol(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 
 // convert functions for options
 int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
+int conv_sndbuf(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_handler(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_bit(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_string(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
@@ -249,6 +251,10 @@ struct option wait_cmds_options[] = {
 	conv_numeric, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
 		DRBD_ ## N ## _DEF ,U,UN  } }
+#define EN_sndbuf(N,U,UN) \
+	conv_sndbuf, show_numeric, numeric_opt_usage, numeric_opt_xml, \
+	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
+		DRBD_ ## N ## _DEF ,U,UN  } }
 #define EH(N,D) \
 	conv_handler, show_handler, handler_opt_usage, handler_opt_xml, \
 	{ .handler_param = { N, ARRY_SIZE(N), \
@@ -280,6 +286,8 @@ struct drbd_cmd commands[] = {
 		 { "on-io-error",'e',	T_on_io_error,	EH(on_error,ON_IO_ERROR) },
 		 { "fencing",'f',	T_fencing,      EH(fencing_n,FENCING) },
 		 { "use-bmbv",'b',	T_use_bmbv,     EB },
+		 { "no-disk-flushes",'i',T_no_disk_flush,EB },
+		 { "no-md-flushes",'m', T_no_md_flush,  EB },
 		 CLOSE_OPTIONS }} }, },
 
 	{"detach", P_detach, F_CONFIG_CMD, {{NULL, NULL}} },
@@ -297,7 +305,7 @@ struct drbd_cmd commands[] = {
 		 { "unplug-watermark",'u',T_unplug_watermark, EN(UNPLUG_WATERMARK,1,NULL) },
 		 { "connect-int",'c',	T_try_connect_int, EN(CONNECT_INT,1,"seconds") },
 		 { "ping-int",'i',	T_ping_int,	   EN(PING_INT,1,"seconds") },
-		 { "sndbuf-size",'S',	T_sndbuf_size,	   EN(SNDBUF_SIZE,1,"bytes") },
+		 { "sndbuf-size",'S',	T_sndbuf_size,	   EN_sndbuf(SNDBUF_SIZE,1,"bytes") },
 		 { "ko-count",'k',	T_ko_count,	   EN(KO_COUNT,1,NULL) },
 		 { "allow-two-primaries",'m',T_two_primaries, EB },
 		 { "cram-hmac-alg",'a',	T_cram_hmac_alg,   ES },
@@ -351,6 +359,8 @@ struct drbd_cmd commands[] = {
 		wait_cmds_options, w_synced_state } } },
 };
 
+#define OTHER_ERROR 900
+
 #define EM(C) [ C - RetCodeBase ]
 
 static const char *error_messages[] = {
@@ -396,7 +406,8 @@ static const char *error_messages[] = {
 	EM(DiskLowerThanOutdated) = "Disk state is lower than outdated",
 	EM(HaveNoDiskConfig) = "Device does not have a disk-config",
 	EM(ProtocolCRequired) = "Protocol C required",
-	EM(VMallocFailed) = "vmalloc() failed. Out of memory?"
+	EM(VMallocFailed) = "vmalloc() failed. Out of memory?",
+	EM(DataOfWrongCurrent) = "Can only attach to the data we lost last (see kernel log).",
 };
 #define MAX_ERROR (sizeof(error_messages)/sizeof(*error_messages))
 const char * error_to_string(int err_no)
@@ -407,8 +418,11 @@ const char * error_to_string(int err_no)
 }
 #undef MAX_ERROR
 
-char* cmdname = 0;
+char *cmdname = NULL; /* "drbdsetup" for reporting in usage etc. */
+char *devname = NULL; /* "/dev/drbd12" for reporting in print_config_error */
+int debug_dump_argv = 0; /* enabled by setting DRBD_DEBUG_DUMP_ARGV in the environment */
 int lock_fd;
+unsigned int cn_idx;
 
 int dump_tag_list(unsigned short *tlc)
 {
@@ -512,24 +526,24 @@ int conv_block_dev(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg
 
 	if ((device_fd = open(arg,O_RDWR))==-1) {
 		PERROR("Can not open device '%s'", arg);
-		return 20;
+		return OTHER_ERROR;
 	}
 
 	if ( (err=fstat(device_fd, &sb)) ) {
 		PERROR("fstat(%s) failed", arg);
-		return 20;
+		return OTHER_ERROR;
 	}
 
 	if(!S_ISBLK(sb.st_mode)) {
 		fprintf(stderr, "%s is not a block device!\n", arg);
-		return 20;
+		return OTHER_ERROR;
 	}
 
 	close(device_fd);
 
 	add_tag(tl,ad->tag,arg,strlen(arg)+1); // include the null byte.
 
-	return 0;
+	return NoError;
 }
 
 int conv_md_idx(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
@@ -542,7 +556,7 @@ int conv_md_idx(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 
 	add_tag(tl,ad->tag,&idx,sizeof(idx));
 
-	return 0;
+	return NoError;
 }
 
 const char* addr_part(const char* s)
@@ -597,7 +611,7 @@ int conv_address(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 
 	add_tag(tl,ad->tag,&addr,addr_len);
 
-	return 0;
+	return NoError;
 }
 
 int conv_protocol(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
@@ -612,12 +626,12 @@ int conv_protocol(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 		prot=DRBD_PROT_C;
 	} else {
 		fprintf(stderr, "'%s' is no valid protocol.\n", arg);
-		return 20;
+		return OTHER_ERROR;
 	}
 
 	add_tag(tl,ad->tag,&prot,sizeof(prot));
 
-	return 0;
+	return NoError;
 }
 
 int conv_bit(struct drbd_option *od, struct drbd_tag_list *tl, char* arg __attribute((unused)))
@@ -626,7 +640,22 @@ int conv_bit(struct drbd_option *od, struct drbd_tag_list *tl, char* arg __attri
 
 	add_tag(tl,od->tag,&bit,sizeof(bit));
 
-	return 0;
+	return NoError;
+}
+
+int conv_sndbuf(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
+{
+	int err = conv_numeric(od, tl, arg);
+	long long l = m_strtoll(arg, 0);
+	char bit = 0;
+
+	if (err != NoError || l != 0)
+		return err;
+	/* this is a mandatory bit,
+	 * to avoid newer userland to configure older modules with
+	 * a sndbuf size of zero, which would lead to Oops. */
+	add_tag(tl, T_auto_sndbuf_size, &bit, sizeof(bit));
+	return NoError;
 }
 
 int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
@@ -644,7 +673,7 @@ int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 		unit[0] = unit_prefix > 1 ? unit_prefix : 0;
 		fprintf(stderr,"%s %s => %llu%s out of range [%llu..%llu]%s\n",
 			od->name, arg, l, unit, min, max, unit);
-		return(20);
+		return OTHER_ERROR;
 	}
 
 	switch(tag_type(od->tag)) {
@@ -658,8 +687,7 @@ int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 	default:
 		fprintf(stderr, "internal error in conv_numeric()\n");
 	}
-	return 0;
-
+	return NoError;
 }
 
 int conv_handler(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
@@ -672,24 +700,32 @@ int conv_handler(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 		if(handler_names[i]==NULL) continue;
 		if(strcmp(arg,handler_names[i])==0) {
 			add_tag(tl,od->tag,&i,sizeof(i));
-			return 0;
+			return NoError;
 		}
 	}
 
-	fprintf(stderr, "Handler not known\n");
-	return 20;
+	fprintf(stderr, "%s-handler '%s' not known\n", od->name, arg);
+	fprintf(stderr, "known %s-handlers:\n", od->name);
+	for (i = 0; i < number_of_handlers; i++) {
+		if (handler_names[i])
+			printf("\t%s\n", handler_names[i]);
+	}
+	return OTHER_ERROR;
 }
 
 int conv_string(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 {
 	add_tag(tl,od->tag,arg,strlen(arg)+1);
 
-	return 0;
+	return NoError;
 }
 
 struct option *	make_longoptions(struct drbd_option* od)
 {
-	static struct option buffer[20];
+	/* room for up to N options,
+	 * plus set-defaults, create-device, and the terminating NULL */
+	const int N = 20;
+	static struct option buffer[N+3];
 	int i=0;
 
 	while(od && od->name) {
@@ -698,8 +734,10 @@ struct option *	make_longoptions(struct drbd_option* od)
 			no_argument : required_argument ;
 		buffer[i].flag = NULL;
 		buffer[i].val = od->short_name;
-		if(i++ == 20) {
+		if (i++ == N) {
+			/* we must not leave this loop with i > N */
 			fprintf(stderr,"buffer in make_longoptions to small.\n");
+			abort();
 		}
 		od++;
 	}
@@ -736,34 +774,45 @@ struct drbd_option *find_opt_by_short_name(struct drbd_option *od, int c)
 	return NULL;
 }
 
-int print_config_error( struct drbd_nl_cfg_reply *reply)
+/* prepends global devname to output (if any) */
+int print_config_error(int err_no)
 {
-	int err_no = reply->ret_code;
 	int rv=0;
 
-	if (err_no == NoError) return rv;
-	if (err_no == SS_Success) return rv;
+	if (err_no == NoError || err_no == SS_Success)
+		return 0;
+	if (err_no == OTHER_ERROR)
+		return 20;
 
 	if ( ( err_no >= AfterLastRetCode || err_no <= RetCodeBase ) &&
 	     ( err_no > SS_CW_NoNeed || err_no < SS_LowerThanOutdated) ) {
-		fprintf(stderr,"Error code %d unknown.\n"
-			"You should updated the drbd userland tools.\n",err_no);
+		fprintf(stderr,"%s: Error code %d unknown.\n"
+			"You should updated the drbd userland tools.\n",
+			devname, err_no);
 		rv = 20;
 	} else {
 		if(err_no > RetCodeBase ) {
-			fprintf(stderr,"Failure: (%d) %s\n",err_no,
-					error_to_string(err_no));
+			fprintf(stderr,"%s: Failure: (%d) %s\n",
+				devname, err_no, error_to_string(err_no));
 			rv = 10;
 		} else if (err_no == SS_UnknownError) {
-			fprintf(stderr,"State change failed: (%d)"
-				"unknown error.\n", err_no);
+			fprintf(stderr,"%s: State change failed: (%d)"
+				"unknown error.\n", devname, err_no);
 			rv = 11;
 		} else if (err_no > SS_TwoPrimaries) {
 			// Ignore SS_Success, SS_NothingToDo, SS_CW_Success...
 		} else {
-			fprintf(stderr,"State change failed: (%d) %s\n",
-				err_no, set_st_err_name(err_no));
-			rv = 11;
+			fprintf(stderr,"%s: State change failed: (%d) %s\n",
+				devname, err_no, set_st_err_name(err_no));
+			if (err_no == SS_NoUpToDateDisk) {
+				/* am Primary, cannot outdate */
+				rv = 17;
+			} else if (err_no == SS_LowerThanOutdated) {
+				/* was inconsistent anyways */
+				rv = 5;
+			} else {
+				rv = 11;
+			}
 		}
 	}
 	return rv;
@@ -771,7 +820,47 @@ int print_config_error( struct drbd_nl_cfg_reply *reply)
 
 #define RCV_SIZE NLMSG_SPACE(sizeof(struct cn_msg)+sizeof(struct drbd_nl_cfg_reply))
 
-int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
+/* cmdname and optind are global variables */
+void warn_unrecognized_option(char **argv)
+{
+	fprintf(stderr, "%s %s: unrecognized option '%s'\n",
+		cmdname, argv[0], argv[optind - 1]);
+}
+
+void warn_missing_required_arg(char **argv)
+{
+	fprintf(stderr, "%s %s: option '%s' requires an argument\n",
+		cmdname, argv[0], argv[optind - 1]);
+}
+
+void warn_print_excess_args(int argc, char **argv, int i)
+{
+	fprintf(stderr, "Ignoring excess arguments:");
+	for (; i < argc; i++)
+		fprintf(stderr, " %s", argv[i]);
+	printf("\n");
+}
+
+void dump_argv(int argc, char **argv, int first_non_option, int n_known_args)
+{
+	int i;
+	if (!debug_dump_argv)
+		return;
+	printf(",-- ARGV dump (optind %d, known_args %d, argc %u):\n",
+		first_non_option, n_known_args, argc);
+	for (i = 0; i < argc; i++) {
+		if (i == 1)
+			puts("-- consumed options:");
+		if (i == first_non_option)
+			puts("-- known args:");
+		if (i == (first_non_option + n_known_args))
+			puts("-- unexpected args:");
+		printf("| %2u: %s\n", i, argv[i]);
+	}
+	printf("`--\n");
+}
+
+int _generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 {
 	char buffer[ RCV_SIZE ];
 	struct drbd_nl_cfg_reply *reply;
@@ -779,8 +868,9 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 	struct drbd_option *od;
 	struct option *lo;
 	struct drbd_tag_list *tl;
-	int c,i=1,rv=0,sk_nl;
+	int c,i=1,rv=NoError,sk_nl;
 	int flags=0;
+	int n_args;
 
 	tl = create_tag_list(4096);
 
@@ -788,39 +878,57 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 		if(argc < i+1) {
 			fprintf(stderr,"Missing argument '%s'\n", ad->name);
 			print_command_usage(cm-commands, "",FULL);
-			rv=20;
-			break;
+			rv = OTHER_ERROR;
+			goto error;
 		}
-		rv |= ad->convert_function(ad,tl,argv[i++]);
-		if(rv) break;
+		rv = ad->convert_function(ad,tl,argv[i++]);
+		if (rv != NoError)
+			goto error;
 		ad++;
 	}
+	n_args = i - 1;
 
 	lo = make_longoptions(cm->cp.options);
 	opterr=0;
-	while( (c=getopt_long(argc,argv,make_optstring(lo,0),lo,0)) != -1 ) {
+	while( (c=getopt_long(argc,argv,make_optstring(lo,':'),lo,0)) != -1 ) {
 		od = find_opt_by_short_name(cm->cp.options,c);
-		if(od) rv |= od->convert_function(od,tl,optarg);
+		if (od)
+			rv = od->convert_function(od,tl,optarg);
 		else {
 			if(c=='(') flags |= DRBD_NL_SET_DEFAULTS;
 			else if(c==')') flags |= DRBD_NL_CREATE_DEVICE;
 			else {
-				fprintf(stderr,
-					"%s: unrecognized option '%s'\n",
-					cmdname, argv[optind-1]);
-				rv=20;
+				if (c == ':') {
+					warn_missing_required_arg(argv);
+					rv = OTHER_ERROR;
+					goto error;
+				}
+				warn_unrecognized_option(argv);
+				rv = OTHER_ERROR;
+				goto error;
 			}
 		}
-		if(rv) break;
+		if (rv != NoError)
+			goto error;
 	}
 
+	/* argc should be cmd + n options + n args;
+	 * if it is more, we did not understand some */
+	if (n_args + optind < argc)
+		warn_print_excess_args(argc, argv, optind + n_args);
+
+	dump_argv(argc, argv, optind, i - 1);
+
 	add_tag(tl,TT_END,NULL,0); // close the tag list
 
-	if(rv == 0) {
+	if(rv == NoError) {
 		//dump_tag_list(tl->tag_list_start);
 		int received;
 		sk_nl = open_cn();
-		if(sk_nl < 0) return 20;
+		if (sk_nl < 0) {
+			rv = OTHER_ERROR;
+			goto error;
+		}
 
 		tl->drbd_p_header->packet_type = cm->packet_id;
 		tl->drbd_p_header->drbd_minor = minor;
@@ -833,14 +941,20 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 		if (received >= 0) {
 			reply = (struct drbd_nl_cfg_reply *)
 				((struct cn_msg *)NLMSG_DATA(buffer))->data;
-			rv = print_config_error(reply);
+			rv = reply->ret_code;
 		}
 	}
+error:
 	free_tag_list(tl);
 
 	return rv;
 }
 
+int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
+{
+	return print_config_error(_generic_config_cmd(cm, minor, argc, argv));
+}
+
 #define ASSERT(exp) if (!(exp)) \
 		fprintf(stderr,"ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__);
 
@@ -999,9 +1113,10 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 	struct drbd_nl_cfg_reply *reply;
 	int sk_nl,rv;
 
-	if(argc > 1) {
-		fprintf(stderr,"Ignoring excess arguments\n");
-	}
+	if (argc > 1)
+		warn_print_excess_args(argc, argv, 1);
+
+	dump_argv(argc, argv, 1, 0);
 
 	tl = create_tag_list(2);
 	add_tag(tl,TT_END,NULL,0); // close the tag list
@@ -1020,7 +1135,7 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 		((struct cn_msg *)NLMSG_DATA(buffer))->data;
 
 	if (reply->ret_code != NoError)
-		return print_config_error(reply);
+		return print_config_error(reply->ret_code);
 
 	rv = cm->gp.show_function(cm,minor,reply->tag_list);
 
@@ -1087,6 +1202,7 @@ int show_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl)
 		       ntohs(addr->sin_port));
 		printf("}\n");
 	}
+	consume_tag_bit(T_auto_sndbuf_size, rtl, &idx); /* consume it, its value has no relevance */
 
 	return 0;
 }
@@ -1095,7 +1211,7 @@ int state_scmd(struct drbd_cmd *cm __attribute((unused)),
 	       int minor __attribute((unused)),
 	       unsigned short *rtl)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1110,7 +1226,7 @@ int cstate_scmd(struct drbd_cmd *cm __attribute((unused)),
 		int minor __attribute((unused)),
 		unsigned short *rtl)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1125,7 +1241,7 @@ int dstate_scmd(struct drbd_cmd *cm __attribute((unused)),
 		int minor __attribute((unused)),
 		unsigned short *rtl)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1180,26 +1296,29 @@ static struct drbd_cmd *find_cmd_by_name(const char* name)
 int down_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 {
 	int rv;
+	int success;
 
 	if(argc > 1) {
 		fprintf(stderr,"Ignoring excess arguments\n");
 	}
 
 	cm = find_cmd_by_name("secondary");
-	rv = cm->function(cm,minor,argc,argv);
-	if( rv ) return rv;
+	rv = _generic_config_cmd(cm, minor, argc, argv); // No error messages
+	if (rv == MinorNotKnown)
+		return 0;
+	success = (rv >= SS_Success && rv < RetCodeBase) || rv == NoError;
+	if (!success)
+		return print_config_error(rv);
 	cm = find_cmd_by_name("disconnect");
 	cm->function(cm,minor,argc,argv);
 	cm = find_cmd_by_name("detach");
-	rv |= cm->function(cm,minor,argc,argv);
-
-	return rv;
+	return cm->function(cm,minor,argc,argv);
 }
 
 int print_state(unsigned int seq, int u __attribute((unused)),
 		struct drbd_nl_cfg_reply *reply)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 	char* str;
 	int synced = 0;
 
@@ -1252,7 +1371,7 @@ int w_connected_state(unsigned int seq __attribute((unused)),
 		      int wait_after_sb,
 		      struct drbd_nl_cfg_reply *reply)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 
 	if(reply->packet_type == P_get_state) {
 		if(consume_tag_int(T_state_i,reply->tag_list,(int*)&state.i)) {
@@ -1268,7 +1387,7 @@ int w_synced_state(unsigned int seq __attribute((unused)),
 		   int wait_after_sb,
 		   struct drbd_nl_cfg_reply *reply)
 {
-	drbd_state_t state;
+	union drbd_state_t state;
 
 	if(reply->packet_type == P_get_state) {
 		if(consume_tag_int(T_state_i,reply->tag_list,(int*)&state.i)) {
@@ -1295,8 +1414,15 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 
 	lo = cm->ep.options;
 
-	while( (c=getopt_long(argc,argv,make_optstring(lo,0),lo,0)) != -1 ) {
+	while( (c=getopt_long(argc,argv,make_optstring(lo,':'),lo,0)) != -1 ) {
 		switch(c) {
+		default:
+		case '?':
+			warn_unrecognized_option(argv);
+			return 20;
+		case ':':
+			warn_missing_required_arg(argv);
+			return 20;
 		case 'u': unfiltered=1; break;
 		case 'a': all_devices=1; break;
 		case 't':
@@ -1327,9 +1453,10 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 		}
 	}
 
-	if(optind > argc) {
-		fprintf(stderr,"Ignoring excess arguments\n");
-	}
+	if (optind < argc)
+		warn_print_excess_args(argc, argv, optind);
+
+	dump_argv(argc, argv, optind, 0);
 
 	tl = create_tag_list(2);
 	add_tag(tl,TT_END,NULL,0); // close the tag list
@@ -1347,6 +1474,17 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 	cn_reply = (struct cn_msg *)NLMSG_DATA(buffer);
 	reply = (struct drbd_nl_cfg_reply *)cn_reply->data;
 	consume_tag_bit(T_use_degraded,reply->tag_list,&rr);
+	if (rr) {
+		if (0 < wfc_timeout &&
+		      (wfc_timeout < degr_wfc_timeout
+				  || degr_wfc_timeout == 0)) {
+			degr_wfc_timeout = wfc_timeout;
+			fprintf(stderr, "degr-wfc-timeout has to be shorter than wfc-timeout\n"
+					"degr-wfc-timeout implicitly set to wfc-timeout (%ds)\n",
+					degr_wfc_timeout);
+		}
+	}
+
 	timeout_ms= 1000 * (  rr ? degr_wfc_timeout : wfc_timeout) - 1;
 
 	// ask for the current state before waiting for state updates...
@@ -1652,7 +1790,7 @@ int open_cn()
 	}
 
 	my_nla.nl_family = AF_NETLINK;
-	my_nla.nl_groups = -1; //CN_IDX_DRBD;
+	my_nla.nl_groups = -1; //cn_idx
 	my_nla.nl_pid = getpid();
 
 	err = bind(sk_nl, (struct sockaddr *)&my_nla, sizeof(my_nla));
@@ -1688,7 +1826,7 @@ void prepare_nl_header(struct nlmsghdr* nl_hdr, int size)
 	nl_hdr->nlmsg_pid = getpid();
 	/* fill the connector header */
 	cn_hdr->id.val = CN_VAL_DRBD;
-	cn_hdr->id.idx = CN_IDX_DRBD;
+	cn_hdr->id.idx = cn_idx;
 	cn_hdr->seq = cn_seq++;
 	get_random_bytes(&cn_hdr->ack,sizeof(cn_hdr->ack));
 	cn_hdr->len = size - sizeof(struct nlmsghdr) - sizeof(struct cn_msg);
@@ -1783,28 +1921,30 @@ void close_cn(int sk_nl)
 
 void ensure_drbd_driver_is_present(void)
 {
-	struct drbd_tag_list *tl;
-	char buffer[4096];
-	int sk_nl, rr;
+	struct stat sb;
+	FILE *cn_idx_file;
+	int err;
 
-	sk_nl = open_cn();
-	/* Might print:
-	   Missing privileges? You should run this as root.
-	   Connector module not loaded? try 'modprobe cn'. */
-	if (sk_nl < 0) exit(20);
+	cn_idx = CN_IDX_DRBD;
+	cn_idx_file = fopen("/sys/module/drbd/parameters/cn_idx", "r");
+	if (cn_idx_file) {
+		fscanf(cn_idx_file, "%u", &cn_idx);
+		fclose(cn_idx_file);
+	}
 
-	tl = create_tag_list(2);
-	add_tag(tl, TT_END, NULL, 0); // close the tag list
+	err = stat("/proc/drbd", &sb);
+	if (!err)
+		return;
 
-	tl->drbd_p_header->packet_type = P_get_state;
-	tl->drbd_p_header->drbd_minor = 0;
-	tl->drbd_p_header->flags = 0;
+	if (err == ENOENT)
+		fprintf(stderr, "DRBD driver appears to be missing\n");
+	else
+		fprintf(stderr, "Could not stat(\"/proc/drbd\"): %m\n");
 
-	rr = call_drbd(sk_nl, tl, (struct nlmsghdr*)buffer, 4096, 500);
-	/* Might print: (after 500ms)
-	   No response from the DRBD driver! Is the module loaded? */
-	close_cn(sk_nl);
-	if (rr == -2) exit(20);
+	fprintf(stderr, "do you need to load the module?\n"
+			"try: modprobe drbd\n");
+
+	exit(20);
 }
 
 int main(int argc, char** argv)
@@ -1830,6 +1970,10 @@ int main(int argc, char** argv)
 		}
 	}
 
+	/* it is enough to set it, value is ignored */
+	if (getenv("DRBD_DEBUG_DUMP_ARGV"))
+		debug_dump_argv = 1;
+
 	if (argc > 1 && (!strcmp(argv[1],"xml"))) {
 		if(argc >= 3) {
 			cmd=find_cmd_by_name(argv[2]);
@@ -1848,6 +1992,8 @@ int main(int argc, char** argv)
 	if(cmd) {
 		lock_fd = dt_lock_drbd(argv[1]);
 		minor=dt_minor_of_dev(argv[1]);
+		/* maybe rather canonicalize, using asprintf? */
+		devname = argv[1];
 		// by passing argc-2, argv+2 the function has the command name
 		// in argv[0], e.g. "syncer"
 		rv = cmd->function(cmd,minor,argc-2,argv+2);
